import pandas as pd
import numpy as np

from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.plotting import figure
from bokeh.models import (
    Button,
    ColumnDataSource,
    # CustomJS,
    DataTable,
    TableColumn,
    Select,
    MultiSelect,
    StringFormatter,  # https://docs.bokeh.org/en/latest/_modules/bokeh/models/widgets/tables.html#DataTable
    Div,
)


### intialize data
df = pd.read_csv("../all_rinko_merged.csv")
df["Residual"] = df["REFOXY_rinko"] - df["CTDRINKO"]
df["Flag"] = 1
df_edited = df.copy()
df_edited["Flag New"] = df["Flag"]
df_edited["Comments"] = ""
ssscc_list = df["SSSCC_rinko"].unique()

# TODO: include bottle numbers

### intialize widgets
button = Button(label="Save flagged data", button_type="success")
# button.js_on_click(CustomJS(args=dict(source=source),
#                             code=open(join(dirname(__file__), "download.js")).read()))
parameter = Select(
    title="Parameter", options=["CTDTMP", "CTDCOND", "SBE43", "RINKO"], value="RINKO"
)
station = Select(title="Station", options=[str(x) for x in ssscc_list], value="00101")
# explanation of flags:
# https://cchdo.github.io/hdo-assets/documentation/manuals/pdf/90_1/chap4.pdf
flag_list = MultiSelect(
    title="Plot data flagged as:",
    value=["1", "2", "3"],
    options=[
        ("1", "1 [Uncalibrated]"),
        ("2", "2 [Acceptable]"),
        ("3", "3 [Questionable]"),
        ("4", "4 [Bad]"),
    ],
)
# returns list of select options, e.g., ['2'] or ['1','2']

source_table = ColumnDataSource(data=dict())
source_table_changes = ColumnDataSource(data=dict())
source_plot_ssscc = ColumnDataSource(data=dict(x=[], y=[]))
source_plot_all = ColumnDataSource(data=dict(x=[], y=[]))

### set up plots
plot_ssscc = figure(
    plot_height=800,
    plot_width=400,
    title="{} vs CTDPRS [Station {}]".format(parameter.value, station.value),
    tools="crosshair,pan,reset,wheel_zoom",
    y_axis_label="Pressure (dbar)",
)
plot_all = figure(
    plot_height=800,
    plot_width=400,
    title="{} vs CTDPRS [Station {}]".format(parameter.value, station.value),
    tools="crosshair,reset,wheel_zoom",
)
plot_ssscc.y_range.flipped = True  # invert y-axis
plot_all.y_range.flipped = True  # invert y-axis
plot_all.scatter(df["REFOXY_rinko"], df["CTDPRS_rinko_ctd"])
plot_ssscc.scatter(
    "x",
    "y",
    fill_color="#999999",
    line_color="#000000",
    size=10,
    line_width=2,
    source=source_plot_ssscc,
)
plot_all.scatter(
    "x",
    "y",
    fill_color="#999999",
    line_color="#000000",
    size=10,
    line_width=2,
    source=source_plot_all,
)

parameter.on_change("value", lambda attr, old, new: update())
station.on_change("value", lambda attr, old, new: update())
flag_list.on_change("value", lambda attr, old, new: update())


def update():

    print("exec update()")

    table_rows = df_edited["SSSCC_rinko"] == int(station.value)
    plot_rows = (df_edited["SSSCC_rinko"] == int(station.value)) & (
        df_edited["Flag New"].isin(flag_list.value)
    )

    # breakpoint()
    current_table = df_edited[table_rows].reset_index()
    current_plot = df_edited[plot_rows]

    # can this be split off into separate updates? might improve speed
    # update_param()
    # update_station()
    # update_plot_flag()
    # update flag()

    source_table.data = {  # this causes update_flag() to execute
        "SSSCC": current_table["SSSCC_rinko"],
        "CTDPRS": current_table["CTDPRS_rinko_ctd"],
        "REFOXY": current_table["REFOXY_rinko"],
        "CTDRINKO": current_table["CTDRINKO"],
        "diff": current_table["Residual"],
        "flag": current_table["Flag New"],
        "comments": current_table["Comments"],
    }
    source_plot_ssscc.data = {
        "x": current_plot["REFOXY_rinko"],
        "y": current_plot["CTDPRS_rinko_ctd"],
    }
    source_plot_all.data = {
        "x": current_plot["REFOXY_rinko"],
        "y": current_plot["CTDPRS_rinko_ctd"],
    }
    plot_ssscc.title.text = "{} vs CTDPRS [Station {}]".format(
        parameter.value, station.value
    )
    plot_all.title.text = "{} vs CTDPRS [Station {}]".format(
        parameter.value, station.value
    )
    plot_ssscc.x_range = plot_all.x_range
    plot_ssscc.y_range = plot_all.y_range


def update_flag():

    print("exec update_flag()")

    # breakpoint()

    df_edited.loc[
        df_edited["SSSCC_rinko"] == source_table.data["SSSCC"].values[0], "Flag New",
    ] = source_table.data["flag"].values
    df_edited.loc[
        df_edited["SSSCC_rinko"] == source_table.data["SSSCC"].values[0], "Comments",
    ] = source_table.data["comments"].values

    edited_rows = (df_edited["Flag"] - df_edited["Flag New"]) != 0

    source_table_changes.data = {
        "SSSCC": df_edited.loc[edited_rows, "SSSCC_rinko"],
        "CTDPRS": df_edited.loc[edited_rows, "CTDPRS_rinko_ctd"],
        "diff": df_edited.loc[edited_rows, "Residual"],
        "flag_old": df_edited.loc[edited_rows, "Flag"],
        "flag_new": df_edited.loc[edited_rows, "Flag New"],
        "comments": df_edited.loc[edited_rows, "Comments"],
    }


def save_data():

    print("Save button clicked...")

    # df_out =


button.on_click(save_data)


columns = [
    TableColumn(
        field="SSSCC",
        title="SSSCC",
        width=40,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="CTDPRS",
        title="CTDPRS",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="REFOXY",
        title="REFOXY",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="CTDRINKO",
        title="CTDRINKO",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="diff",
        title="Residual",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="flag",
        title="Flag",
        width=20,
        formatter=StringFormatter(text_align="center", font_style="bold"),
    ),
    TableColumn(
        field="comments",
        title="Comments",
        width=100,
        formatter=StringFormatter(text_align="left"),
    ),
]
columns_changed = [
    TableColumn(
        field="SSSCC",
        title="SSSCC",
        width=40,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="CTDPRS",
        title="CTDPRS",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="diff",
        title="Residual",
        width=80,
        formatter=StringFormatter(text_align="right"),
    ),
    TableColumn(
        field="flag_old",
        title="Old",
        width=20,
        formatter=StringFormatter(text_align="center", font_style="bold"),
    ),
    TableColumn(
        field="flag_new",
        title="New",
        width=20,
        formatter=StringFormatter(
            text_align="center", font_style="bold", text_color="red"
        ),
    ),
    TableColumn(
        field="comments",
        title="Comments",
        width=100,
        formatter=StringFormatter(text_align="left"),
    ),
]
data_table = DataTable(
    source=source_table,
    columns=columns,
    index_width=20,
    width=480 + 20,  # sum of col widths + idx width
    height=600,
    editable=True,
    fit_columns=True,
    sortable=False,
)
data_table_changed = DataTable(
    source=source_table_changes,
    columns=columns_changed,
    index_width=20,
    width=480 + 20,  # sum of col widths + idx width
    height=200,
    editable=False,
    fit_columns=True,
    sortable=False,
)
data_table_title = Div(text="""<b>All Station Data:</b>""", width=200, height=15)
data_table_changed_title = Div(text="""<b>Edited Data:</b>""", width=200, height=15)

# run update() when user selects new column (may indicate new flag value)
# source_table.selected.on_change("indices", lambda attr, old, new: update_flag())
source_table.on_change("data", lambda attr, old, new: update_flag())

# TODO: highlight scatterpoint based on selected row
# source_table.selected.on_change("indices", lambda attr, old, new: update_highlight())
# source_table.selected.indices -> could likely be used to highlight point

controls = column(parameter, station, flag_list, button, width=170)
tables = column(
    data_table_title, data_table, data_table_changed_title, data_table_changed
)

curdoc().add_root(row(controls, tables, plot_ssscc, plot_all))
curdoc().title = "CTDO Data Flagging Tool"

update()



# ---
# jupyter:
#   jupytext:
#     cell_markers: '"""'
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.9.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
"""
# Introducing Jupyter Notebooks
First, set up the environment:
"""

# %%
import matplotlib
import matplotlib.pyplot as pl
import numpy as np

try:
    from IPython import get_ipython

    get_ipython().run_line_magic("matplotlib", "inline")
except AttributeError:
    print("Magic function can only be used in IPython environment")
    matplotlib.use("Agg")

pl.rcParams["figure.figsize"] = [15, 8]

# %% [markdown]
"""
Then, define a function that creates a pretty graph:
"""

# %%
def SineAndCosineWaves():
    # Get a large number of X values for a nice smooth curve. Using Pi as np.sin requires radians...
    x = np.linspace(0, 2 * np.pi, 180)
    # Convert radians to degrees to make for a meaningful X axis (1 radian = 57.29* degrees)
    xdeg = 57.29577951308232 * np.array(x)
    # Calculate the sine of each value of X
    y = np.sin(x)
    # Calculate the cosine of each value of X
    z = np.cos(x)
    # Plot the sine wave in blue, using degrees rather than radians on the X axis
    pl.plot(xdeg, y, color="blue", label="Sine wave")
    # Plot the cos wave in green, using degrees rather than radians on the X axis
    pl.plot(xdeg, z, color="green", label="Cosine wave")
    pl.xlabel("Degrees")
    # More sensible X axis values
    pl.xticks(np.arange(0, 361, 45))
    pl.legend()
    pl.show()


# %% [markdown]
"""
Finally, call that function to display the graph:
"""

# %%
SineAndCosineWaves()



#! /usr/bin/env python
# remove and streamline imports below later
import argparse
import os
import sys

from . import convert as cnv
from . import sbe_reader as sbe_reader

DEBUG = False

# File extension to use for output files (csv-formatted)
FILE_EXT = "csv"

# File extension to use for raw output
RAW_SUFFIX = "_raw"

# File extension to use for converted output
CONVERTED_SUFFIX = "_cnv"

PKL_EXT = "pkl"


def debugPrint(*args, **kwargs):
    if DEBUG:
        errPrint(*args, **kwargs)


def errPrint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


# -------------------------------------------------------------------------------------
# Main function of the script should it be run as a stand-alone utility.
# -------------------------------------------------------------------------------------
def main(argv):
    # MK: depreciated 04/27/20, use ctdcal.convert.hex_to_ctd instead

    parser = argparse.ArgumentParser(
        description="Convert SBE raw data to a converted, csv-formatted text file"
    )
    parser.add_argument(
        "hexFile", metavar="hex_file", help="the .hex data file to process"
    )
    parser.add_argument(
        "xmlconFile", metavar="XMLCON_file", help="the .XMLCON data file to process"
    )

    # debug messages
    # parser.add_argument('-d', '--debug', action='store_true', help='display debug messages')

    # raw output
    # parser.add_argument('-r', '--raw', action='store_true', help='return the raw data values')

    # output directory
    parser.add_argument(
        "-o", metavar="dest_dir", dest="outDir", help="location to save output files"
    )

    # Process Command-line args
    args = parser.parse_args()
    # if args.debug:
    #     global DEBUG
    #     DEBUG = True
    #     debugPrint("Running in debug mode")

    # Verify hex file exists
    if not os.path.isfile(args.hexFile):
        errPrint("ERROR: Input hex file:", args.hexFile, "not found\n")
        sys.exit(1)

    # Verify xmlcon file exists
    if not os.path.isfile(args.xmlconFile):
        errPrint("ERROR: Input xmlcon file:", args.xmlconFile, "not found\n")
        sys.exit(1)

    # Set the default output directory to be the same directory as the hex file
    outputDir = os.path.dirname(args.hexFile)

    # Used later for building output file names
    filename_ext = os.path.basename(args.hexFile)  # original filename with ext
    filename_base = os.path.splitext(filename_ext)[0]  # original filename w/o ext

    # Parse the input files
    debugPrint("Parsing", args.hexFile, "and", args.xmlconFile + "... ", end="")
    sbeReader = sbe_reader.SBEReader.from_paths(args.hexFile, args.xmlconFile)
    debugPrint("Success!")

    # Build Output Directory exists
    if args.outDir:
        if os.path.isdir(args.outDir):
            outputDir = args.outDir
        else:
            debugPrint("Creating output directory...", end="")
            try:
                os.mkdir(args.outDir)
            except OSError:
                errPrint("ERROR: Could not create output directory:", args.outDir)
                sys.exit(1)
            else:
                outputDir = args.outDir
                debugPrint("Success!")

    # # Save the raw scans as csv
    # if args.raw:
    #
    #     debugPrint('Building raw dataset... ', end='')
    #
    #     # Retrieve parsed scans
    #     rawData = sbeReader.parsed_scans()
    #
    #     # Convert raw data to dataframe
    #     raw_df = pd.DataFrame(rawData)
    #     raw_df.index.name = 'index'
    #     raw_df = raw_df.apply(pd.to_numeric, errors="ignore")
    #
    #     debugPrint('Success!')
    #
    #     rawfileName = str(filename_base + RAW_SUFFIX + '.' + FILE_EXT)
    #     rawfilePath = os.path.join(outputDir, rawfileName)
    #
    #     debugPrint('Saving raw data to:', rawfilePath + '... ', end='')
    #     try:
    #         raw_df.to_csv(rawfilePath)
    #     except:
    #         errPrint('ERROR: Could not save raw data to file')
    #     else:
    #         debugPrint('Success!')

    # debugPrint("Converting raw scans to scientific units... ")
    converted_df = cnv.convertFromSBEReader(sbeReader, False)

    # convertedfileName  = filename_base + CONVERTED_SUFFIX + '.' + FILE_EXT
    # convertedfilePath = os.path.join(outputDir, convertedfileName)

    ### Test pickle file conversion
    pickle_file_name = filename_base + "." + PKL_EXT
    pickle_file_path = os.path.join(outputDir, pickle_file_name)
    converted_df.to_pickle(pickle_file_path)

    # debugPrint('Saving converted data to:', convertedfilePath + '... ', end='')
    # if cnv.saveConvertedDataToFile(converted_df, convertedfilePath, False):
    #     debugPrint('Success!')
    # else:
    #     errPrint('ERROR: Could not save converted data to file')


# -------------------------------------------------------------------------------------
# Required python code for running the script as a stand-alone utility
# -------------------------------------------------------------------------------------
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
import os
import pathlib

cruise = dict(
    cruise_title="GO-SHIP NBP1802",
    cruise_name="nbp1802",
    cruisedb="nbp1802",
    vessel="R/V Palmer",
    ship_code="3206",
    expocode="320620180309",
    chief_sci="Alison Macdonald",
    sectionid="S04P",
    start_date="2018-03-09 14:00:00",
    start_port="Hobart, Tasmania, Australia",
    start_latlon="42.8821 S 147.3272 W",
    end_date="2018-05-14 10:00:00",
    end_port="Punta Arenas, Chile",
    end_latlon="53.1638 S 70.9171 W",
)

ctd_processing_dir = dict(
    raw_data_directory="data/raw/",
    time_data_directory="data/time/",
    pressure_data_directory="data/pressure/",
    fit_data_directory="data/fit/",
    converted_directory="data/converted/",
    log_directory="data/logs/",
    salt_directory="data/salt/",
    oxygen_directory="data/oxygen/",
    bottle_directory="data/bottle/",
    reft_directory="data/reft/",
    o2flask_file="data/oxygen/o2flasks.vol",
    pressure_log="data/logs/ondeck_pressure.csv",
    hex_prefix="data/raw/",
    hex_postfix=".hex",
    xml_prefix="data/raw/",
    xml_postfix=".XMLCON",
)

ctd_processing_constants = dict(
    conductivity_start=20.0,
    pressure_interval=2.0,
    sample_rate=4,
    roll_filter_time=15,
    TC_side=2,
    roll_filter=1,
    hysteresis_1=-0.033,
    hysteresis_2=5000,
    hysteresis_3=1450,
    TC_primary_align=0.0,
    TC_secondary_align=0.0,
    DO_align=2.0,
    ctd_serial=1281,
)


bottle_inputs = dict(
    p="CTDPRS",  # Pressure column
    t="CTDTMP2",  # CTD Temperature
    t1="CTDTMP1",  # CTD Temperature
    t2="CTDTMP2",  # CTD Temperature
    c="CTDCOND1",  # CTD Conductivity
    c1="CTDCOND1",  # CTD Conductivity 1
    c2="CTDCOND2",  # CTD Conductivity 2
    salt="CTDSAL",  # CTD Salinity
    dopl="CTDOXY1",  # CTD Oxygen (ml/L)
    dov="CTDOXYVOLTS",  # CTD Oxygen volts
    xmis="CTDXMISS",  # Transmissometer column
    fluor="FLUOR",  # Fluorometer column
    backscatter="CTDBACKSCATTER",  # Backscatter column
    rinko_oxy="FREE1",  # Rinko Oxygen column (additional column)
    rinko_temp="FREE3",  # Rinko Temp column (additional column)
    alt="ALT",  # Altimeter
    ref_par="REF_PAR",
    lat="GPSLAT",  # Latitutde
    lon="GPSLON",  # Longitude
    scan_datetime="scan_datetime",  # Scan time
    nmea_datetime="nmea_datetime",
    btl_num="btl_fire_num",  # Bottle fire number/bottle index column
    btl_fire="btl_fire",  # Bottle fire check (0/1)
    pump_on="pump_on",  # Pump on check (0/1)
    time_zone="UTC",  # Time Zone
    #    T90 = 'T90', # Reference Temp
    reft="REFTMP",
    bath_temp="BathTEMP",  # Salinometer bath temp
    cond_ratio="CRavg",  # Salinometer conductivity ratio
    btl_cond="BTLCOND",  # Salinometer bottle conductivity
    btl_oxy="OXYGEN",  # Bottle Oxygen
)

btl_input_array = [
    bottle_inputs["p"],
    bottle_inputs["t1"],
    bottle_inputs["t2"],
    bottle_inputs["c1"],
    bottle_inputs["c2"],
    bottle_inputs["dov"],
    bottle_inputs["dopl"],
    bottle_inputs["xmis"],
    bottle_inputs["alt"],
    bottle_inputs["rinko_oxy"],
    bottle_inputs["rinko_temp"],
    bottle_inputs["lat"],
    bottle_inputs["lon"],
    bottle_inputs["pump_on"],
    bottle_inputs["btl_fire"],
    bottle_inputs["scan_datetime"],
    bottle_inputs["btl_num"],
    bottle_inputs["fluor"],
]  # bottle_inputs['fluor']

# Dimensioned/flagged bottle outputs:
# Form = ['PARAMETER', 'PARAMETER_FLAG_NAME', 'PARAMETER_UNITS','PARAMETER_FLAG_UNITS']
btl_outputs = dict(
    btl_nbr=["BTLNBR", "BTLNBR_FLAG_W", "", ""],
    press=["CTDPRS", "CTDPRS_FLAG_W", "DBAR", ""],
    temp=["CTDTMP", "CTDTMP_FLAG_W", "ITS-90", ""],
    ref_tmp=["REFTMP", "REFTMP_FLAG_W", "ITS-90", ""],
    salt=["CTDSAL", "CTDSAL_FLAG_W", "PSS-78", ""],
    slnty=["SALNTY", "SALNTY_FLAG_W", "PSS-78", ""],
    ctd_oxy=["CTDOXY", "CTDOXY_FLAG_W", "UMOL/KG", ""],
    oxy=["OXYGEN", "OXYGEN_FLAG_W", "UMOL/KG", ""],
    rinko=["CTDRINKO", "CTDRINKO_FLAG_W", "UMOL/KG", ""],
    xmiss=["CTDXMISS", "CTDXMISS_FLAG_W", "0-5VDC", ""],
    fluor=["CTDFLUOR", "CTDFLUOR_FLAG_W", "0-5VDC", ""],
    # bbp = ['CTDBBP700RAW', 'CTDBBP700RAW_FLAG_W', '0-5VDC', ''],
)
# Undimensioned/unflagged bottle outputs:
# Form = ['PARAMETER','PARAMETER NOTE/UNIT'] parameter note/unit is what will be displayed underneath the parameter name in the exchange file
btl_und_outputs = dict(
    expocd=["EXPOCODE", ""],
    sectid=["SECT_ID", ""],
    stnnbr=["STNNBR", ""],
    castno=["CASTNO", ""],
    sampno=["SAMPNO", ""],
    date=["DATE", ""],
    time=["TIME", ""],
    lat=["LATITUDE", ""],
    lon=["LONGITUDE", ""],
    dep=["DEPTH", ""],
)
btl_column_names = []
btl_column_units = []
btl_flagged_params = []

for i in range(len(btl_und_outputs)):
    param = list(btl_und_outputs.keys())[i]
    param_list = btl_und_outputs[param]
    btl_column_names.append(param_list[0])
    btl_column_units.append(param_list[1])

for i in range(len(btl_outputs)):
    param = list(btl_outputs.keys())[i]
    param_list = btl_outputs[param]
    btl_column_names.append(param_list[0])
    btl_column_names.append(param_list[1])
    btl_column_units.append(param_list[2])
    btl_column_units.append(param_list[3])
    btl_flagged_params.append(param_list[0])

btl_series_output = dict(
    btl_column_names=btl_column_names, btl_column_units=btl_column_units
)

ctd_inputs = dict(
    p="CTDPRS",  # Pressure column
    t="CTDTMP2",  # CTD Temp
    t1="CTDTMP1",  # CTD Temp 1
    t2="CTDTMP2",  # CTD Temp 2
    reft="REFTMP",  # Reference Temp
    c="CTDCOND2",  # CTD Conductivity
    c1="CTDCOND1",  # CTD Conductivity 1
    c2="CTDCOND2",  # CTD Conductivity 2
    salt="CTDSAL",  # CTD Salinity
    btl_salt="SALNTY",  # Bottle Salinity
    dov="CTDOXYVOLTS",  # CTD Oxygen Volts
    dopl="CTDOXY1",  # CTD Oxygen (ml/L)
    dopkg="CTDOXY1",  # CTD Oxygem (umol/KG)
    xmis="CTDXMISS",  # Transmissometer
    fluor="CTDFLUOR",  # Fluorometer
    backscatter="CTDBACKSCATTER",  # Backscatter
    rinko_oxy="FREE1",  # Rinko Oxygen column (additional column)
    rinko_tmp="FREE2",  # Rinko Temperature column (additional column)
    pump_on="pump_on",  # Pump on check (0/1)
    alt="ALT",  # Altimeter
    scan_datetime="scan_datetime",  # Scan time
    datetime="GPSEPOCHTIME",  # Time
    lat="GPSLAT",  # Latitude
    lon="GPSLON",  # Longitude
)
ctd_outputs = dict(
    press=["CTDPRS", "CTDPRS_FLAG_W", "DBAR", ""],
    temp=["CTDTMP", "CTDTMP_FLAG_W", "ITS-90", ""],
    salt=["CTDSAL", "CTDSAL_FLAG_W", "PSS-78", ""],
    doxy=["CTDOXY", "CTDOXY_FLAG_W", "UMOL/KG", ""],
    rinko=["CTDRINKO", "CTDRINKO_FLAG_W", "0-5VDC", ""],
    xmiss=["CTDXMISS", "CTDXMISS_FLAG_W", "0-5VDC", ""],
    fluor=["CTDFLUOR", "CTDFLUOR_FLAG_W", "0-5VDC", ""],
    # bbp = ['CTDBBP700RAW', 'CTDBBP700RAW_FLAG_W', '0-5VDC', ''],
)

ctd_input_array = [
    ctd_inputs["p"],
    ctd_inputs["t1"],
    ctd_inputs["t2"],
    ctd_inputs["c1"],
    ctd_inputs["c2"],
    ctd_inputs["dov"],
    ctd_inputs["dopl"],
    ctd_inputs["xmis"],
    ctd_inputs["alt"],
    ctd_inputs["rinko_oxy"],
    ctd_inputs["rinko_tmp"],
    ctd_inputs["lat"],
    ctd_inputs["lon"],
    ctd_inputs["pump_on"],
    ctd_inputs["scan_datetime"],
]  # ,ctd_inputs['fluor']

# ctd_input_array = [ctd_inputs['p'], ctd_inputs['t1'], ctd_inputs['t2'], ctd_inputs['c1'], ctd_inputs['c2'], ctd_inputs['dov'], ctd_inputs['dopl'],
#                   ctd_inputs['xmis'], ctd_inputs['alt'], ctd_inputs['fluor'], ctd_inputs['backscatter'], ctd_inputs['rinko_oxy'],ctd_inputs['rinko_tmp'],
#                   ctd_inputs['lat'], ctd_inputs['lon'], ctd_inputs['pump_on'], ctd_inputs['scan_datetime']]

time_series_output = dict(
    data_output=[
        "CTDPRS_DBAR",
        "CTDTMP1_ITS90",
        "CTDTMP2_ITS90",
        "CTDCOND1_MSPCM",
        "CTDCOND2_MSPCM",
        "CTDSAL_PSU",
        "CTDOXY1_MLPL",
        "CTDOXYVOLTS",
        "CTDXMISS_05VDC",
        "FLUOR_05VDC",
        "scan_datetime",
        "LATITUDE",
        "LONGITUDE",
    ],
    column_name=[
        ctd_outputs["press"][0],
        "CTDTMP1",
        "CTDTMP2",
        "CTDCOND1",
        "CTDCOND2",
        "CTDSAL",
        "CTDOXY",
        "CTDOXYVOLTS",
        "TRANS",
        "FLUOR",
        "GPSEPOCHTIME",
        "GPSLAT",
        "GPSLON",
    ],
    column_units=[
        "DBAR",
        "ITS90 degC",
        "ITS90 degC",
        "MS/CM",
        "MS/CM",
        "PSS78 PSU",
        "ML/L",
        "0-5VDC",
        "0-5VDC",
        "0-5VDC",
        "SECONDS",
        "-S",
        "-W",
    ],
    format=[
        "%8.1f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%10.4f",
        "%d",
        "%10.5f",
        "%10.5f\n",
    ],
)

ctd_column_names = []
ctd_column_units = []
for i in range(len(ctd_outputs)):
    param = list(ctd_outputs.keys())[i]
    param_list = ctd_outputs[param]
    ctd_column_names.append(param_list[0])
    ctd_column_names.append(param_list[1])
    ctd_column_units.append(param_list[2])
    ctd_column_units.append(param_list[3])

pressure_series_output = dict(
    column_names=ctd_column_names, column_units=ctd_column_units
)


input_array = [
    ctd_inputs["p"],
    ctd_inputs["t"],
    ctd_inputs["salt"],
    ctd_inputs["dopl"],
    ctd_inputs["dov"],
    ctd_inputs["xmis"],
    ctd_inputs["fluor"],
]
# Create necessary files:
for key, val in ctd_processing_dir.items():
    if key != "o2flask_file" and key != "pressure_log":
        pathlib.Path(val).mkdir(exist_ok=True)
    else:
        continue

# Generate ssscc from directory
ssscc = os.listdir(ctd_processing_dir["raw_data_directory"])
xml_list = []
hex_list = []
for s in ssscc:
    if ".XMLCON" in s:
        xml_list.append(s[:-7])
    elif ".hex" in s:
        hex_list.append(s[:-4])
    else:
        pass
ssscc = list(set(xml_list) & set(hex_list))
ssscc.sort()


do_primary = 1  # 0
do_secondary = 1  # 0



import pandas as pd
import glob
import pickle
import gsw

from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.plotting import figure
from bokeh.models import (
    Button,
    ColumnDataSource,
    DataTable,
    TableColumn,
    Select,
    MultiSelect,
    StringFormatter,  # https://docs.bokeh.org/en/latest/_modules/bokeh/models/widgets/tables.html#DataTable
    Div,
    TextInput,
    BoxSelectTool,
)

# TODO: abstract parts of this to a separate file
# TODO: following above, make parts reusable?

# load continuous CTD data and make into a dict (only ~20MB)
file_list = sorted(glob.glob("../../data/pressure/*ct1.csv"))
ssscc_list = [ssscc.strip("../../data/pressure/")[:5] for ssscc in file_list]
ctd_data = []
for f in file_list:
    print(f"Loading {f}")
    df = pd.read_csv(f, header=12, skiprows=[13], skipfooter=1, engine="python")
    df["SSSCC"] = f.strip("../../data/pressure/")[:5]
    ctd_data.append(df)
ctd_data = pd.concat(ctd_data, axis=0, sort=False)

# load bottle trip file
file_list = sorted(glob.glob("../../data/bottle/*.pkl"))
ssscc_list = [ssscc.strip("../../data/bottle/")[:5] for ssscc in file_list]
upcast_data = []
for f in file_list:
    with open(f, "rb") as x:
        df = pickle.load(x)
        df["SSSCC"] = f.strip("../../data/bottle/")[:5]
        # change to secondary if that is what's used
        upcast_data.append(df[["SSSCC", "CTDCOND1", "CTDTMP1", "CTDPRS"]])
upcast_data = pd.concat(upcast_data, axis=0, sort=False)
upcast_data["CTDSAL"] = gsw.SP_from_C(
    upcast_data["CTDCOND1"], upcast_data["CTDTMP1"], upcast_data["CTDPRS"]
)

# load salt file (adapted from compare_salinities.ipynb)
file_list = sorted(glob.glob("../../data/salt/*.csv"))
ssscc_list = [ssscc.strip("../../data/salt/")[:5] for ssscc in file_list]
salt_data = []
for f in file_list:
    df = pd.read_csv(f, usecols=["STNNBR", "CASTNO", "SAMPNO", "SALNTY"])
    df["SSSCC"] = f.strip("../../data/salt/")[:5]
    salt_data.append(df)
salt_data = pd.concat(salt_data, axis=0, sort=False)
salt_data["SALNTY"] = salt_data["SALNTY"].round(4)
if "SALNTY_FLAG_W" not in salt_data.columns:
    salt_data["SALNTY_FLAG_W"] = 2

# load ctd btl data
df_ctd_btl = pd.read_csv(
    "../../data/scratch_folder/ctd_to_bottle.csv",
    skiprows=[1],
    skipfooter=1,
    engine="python",
)
df_btl_all = pd.merge(df_ctd_btl, salt_data, on=["STNNBR", "CASTNO", "SAMPNO"])
btl_data = df_btl_all.loc[
    :,
    [
        "SSSCC",
        "SAMPNO",
        "CTDPRS",
        "CTDTMP",
        "REFTMP",
        "CTDSAL",
        "SALNTY",
        "SALNTY_FLAG_W",
    ],
]
btl_data["Residual"] = btl_data["SALNTY"] - btl_data["CTDSAL"]
btl_data[["CTDPRS", "Residual"]] = btl_data[["CTDPRS", "Residual"]].round(4)
btl_data["Comments"] = ""
btl_data["New Flag"] = btl_data["SALNTY_FLAG_W"].copy()

# update with old handcoded flags if file exists
handcoded_file = "salt_flags_handcoded.csv"
if glob.glob(handcoded_file):
    handcodes = pd.read_csv(handcoded_file, dtype={"SSSCC": str}, keep_default_na=False)
    handcodes = handcodes.rename(columns={"salinity_flag": "New Flag"}).drop(
        columns="diff"
    )
    # there's gotta be a better way... but this is good enough for now
    btl_data = btl_data.merge(handcodes, on=["SSSCC", "SAMPNO"], how="left")
    merge_rows = ~btl_data["New Flag_y"].isnull() | ~btl_data["Comments_y"].isnull()
    btl_data.loc[merge_rows, "New Flag_x"] = btl_data.loc[merge_rows, "New Flag_y"]
    btl_data.loc[merge_rows, "Comments_x"] = btl_data.loc[merge_rows, "Comments_y"]
    btl_data = btl_data.rename(
        columns={"New Flag_x": "New Flag", "Comments_x": "Comments"}
    ).drop(columns=["New Flag_y", "Comments_y"])

# intialize widgets
save_button = Button(label="Save flagged data", button_type="success")
parameter = Select(title="Parameter", options=["CTDSAL", "CTDTMP"], value="CTDSAL")
ref_param = Select(title="Reference", options=["SALNTY"], value="SALNTY")
# ref_param.options = ["foo","bar"]  # can dynamically change dropdowns
station = Select(title="Station", options=ssscc_list, value=ssscc_list[0])
# explanation of flags:
# https://cchdo.github.io/hdo-assets/documentation/manuals/pdf/90_1/chap4.pdf
flag_list = MultiSelect(
    title="Plot data flagged as:",
    value=["1", "2", "3"],
    options=[
        ("1", "1 [Uncalibrated]"),
        ("2", "2 [Acceptable]"),
        ("3", "3 [Questionable]"),
        ("4", "4 [Bad]"),
    ],
)
# returns list of select options, e.g., ['2'] or ['1','2']
flag_input = Select(
    title="Flag:",
    options=[
        ("1", "1 [Uncalibrated]"),
        ("2", "2 [Acceptable]"),
        ("3", "3 [Questionable]"),
        ("4", "4 [Bad]"),
    ],
    value="3",
)
comment_box = TextInput(value="", title="Comment:")

# button_type: default, primary, success, warning or danger
flag_button = Button(label="Apply to selected", button_type="primary")
comment_button = Button(label="Apply to selected", button_type="warning")

vspace = Div(text=""" """, width=200, height=65)
bulk_flag_text = Div(
    text="""<br><br>
    <b>Bulk Bottle Flagging:</b><br>
    Select multiple bottles using the table
    (with shift/control) or the 'Box Select' tool on the plot.""",
    width=150,
    height=135,
)

# set up datasources
src_table = ColumnDataSource(data=dict())
src_table_changes = ColumnDataSource(data=dict())
src_plot_trace = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_ctd = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_upcast = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_btl = ColumnDataSource(data=dict(x=[], y=[]))

# set up plots
fig = figure(
    plot_height=800,
    plot_width=400,
    title="{} vs CTDPRS [Station {}]".format(parameter.value, station.value),
    tools="pan,box_zoom,wheel_zoom,box_select,reset",
    y_axis_label="Pressure (dbar)",
)
fig.line(
    "x",
    "y",
    line_color="#000000",
    line_width=2,
    source=src_plot_trace,
    legend_label="CTD Trace",
)
btl_sal = fig.asterisk(
    "x",
    "y",
    size=12,
    line_width=1.5,
    color="#0033CC",
    source=src_plot_btl,
    legend_label="Bottle sample",
)
ctd_sal = fig.circle(
    "x",
    "y",
    size=7,
    color="#BB0000",
    source=src_plot_ctd,
    legend_label="Downcast CTD sample",
)
upcast_sal = fig.triangle(
    "x",
    "y",
    size=7,
    color="#00BB00",
    source=src_plot_upcast,
    legend_label="Upcast CTD sample",
)
fig.select(BoxSelectTool).select_every_mousemove = False
fig.y_range.flipped = True  # invert y-axis
fig.legend.location = "bottom_right"
fig.legend.border_line_width = 3
fig.legend.border_line_alpha = 1
btl_sal.nonselection_glyph.line_alpha = 0.2
ctd_sal.nonselection_glyph.fill_alpha = 1  # makes CTDSAL *not* change on select
upcast_sal.nonselection_glyph.fill_alpha = 1  # makes CTDSAL *not* change on select

# define callback functions


def update_selectors():

    print("exec update_selectors()")
    ctd_rows = ctd_data["SSSCC"] == station.value
    table_rows = btl_data["SSSCC"] == station.value
    btl_rows = (btl_data["New Flag"].isin([int(v) for v in flag_list.value])) & (
        btl_data["SSSCC"] == station.value
    )
    # breakpoint()

    # update table data
    current_table = btl_data[table_rows].reset_index()
    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "CTDSAL": current_table["CTDSAL"],
        "SALNTY": current_table["SALNTY"],
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
    }

    # update plot data
    src_plot_trace.data = {
        "x": ctd_data.loc[ctd_rows, parameter.value],
        "y": ctd_data.loc[ctd_rows, "CTDPRS"],
    }
    src_plot_ctd.data = {
        "x": btl_data.loc[table_rows, parameter.value],
        "y": btl_data.loc[table_rows, "CTDPRS"],
    }
    src_plot_upcast.data = {
        "x": upcast_data.loc[upcast_data["SSSCC"] == station.value, "CTDSAL"],
        "y": upcast_data.loc[upcast_data["SSSCC"] == station.value, "CTDPRS"],
    }
    src_plot_btl.data = {
        "x": btl_data.loc[btl_rows, "SALNTY"],
        "y": btl_data.loc[btl_rows, "CTDPRS"],
    }

    # update plot labels/axlims
    fig.title.text = "{} vs CTDPRS [Station {}]".format(parameter.value, station.value)
    fig.xaxis.axis_label = parameter.value

    # deselect all datapoints
    btl_sal.data_source.selected.indices = []
    src_table.selected.indices = []


def edit_flag():

    print("exec edit_flag()")

    btl_data.loc[
        btl_data["SSSCC"] == src_table.data["SSSCC"].values[0], "New Flag",
    ] = src_table.data["flag"].values
    btl_data.loc[
        btl_data["SSSCC"] == src_table.data["SSSCC"].values[0], "Comments",
    ] = src_table.data["Comments"].values

    edited_rows = (btl_data["New Flag"].isin([3, 4])) | (btl_data["Comments"] != "")

    src_table_changes.data = {
        "SSSCC": btl_data.loc[edited_rows, "SSSCC"],
        "SAMPNO": btl_data.loc[edited_rows, "SAMPNO"],
        "diff": btl_data.loc[edited_rows, "Residual"],
        "flag_old": btl_data.loc[edited_rows, "SALNTY_FLAG_W"],
        "flag_new": btl_data.loc[edited_rows, "New Flag"],
        "Comments": btl_data.loc[edited_rows, "Comments"],
    }


def apply_flag():

    print("Applying flags")

    table_rows = btl_data["SSSCC"] == station.value
    selected_rows = src_table.selected.indices

    # update table data
    current_table = btl_data[table_rows].reset_index()
    current_table.loc[selected_rows, "New Flag"] = int(flag_input.value)
    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "CTDSAL": current_table["CTDSAL"],
        "SALNTY": current_table["SALNTY"],
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
    }


def apply_comment():

    print("Applying Comments")

    table_rows = btl_data["SSSCC"] == station.value
    selected_rows = src_table.selected.indices

    # update table data
    current_table = btl_data[table_rows].reset_index()
    current_table.loc[selected_rows, "Comments"] = comment_box.value
    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "CTDSAL": current_table["CTDSAL"],
        "SALNTY": current_table["SALNTY"],
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
    }


def save_data():

    print("Saving flagged data...")

    # get data from table
    df_out = pd.DataFrame.from_dict(src_table_changes.data)

    # minor changes to columns/names/etc.
    df_out = df_out.rename(columns={"flag_new": "salinity_flag"}).drop(
        columns="flag_old"
    )

    # save it
    df_out.to_csv("salt_flags_handcoded.csv", index=None)


def selected_from_plot(attr, old, new):

    src_table.selected.indices = new


def selected_from_table(attr, old, new):

    btl_sal.data_source.selected.indices = new


# set up change callbacks
parameter.on_change("value", lambda attr, old, new: update_selectors())
station.on_change("value", lambda attr, old, new: update_selectors())
flag_list.on_change("value", lambda attr, old, new: update_selectors())
flag_button.on_click(apply_flag)
comment_button.on_click(apply_comment)
save_button.on_click(save_data)
src_table.on_change("data", lambda attr, old, new: edit_flag())
src_table.selected.on_change("indices", selected_from_table)
btl_sal.data_source.selected.on_change("indices", selected_from_plot)


# build data tables
columns = []
fields = ["SSSCC", "SAMPNO", "CTDPRS", "CTDSAL", "SALNTY", "diff", "flag", "Comments"]
titles = ["SSSCC", "Bottle", "CTDPRS", "CTDSAL", "SALNTY", "Residual", "Flag", "Comments"]
widths = [40, 20, 75, 65, 65, 65, 15, 135]
for (field, title, width) in zip(fields, titles, widths):
    if field == "flag":
        strfmt_in = {"text_align": "center", "font_style": "bold"}
    elif field == "Comments":
        strfmt_in = {}
    else:
        strfmt_in = {"text_align": "right"}
    columns.append(TableColumn(
        field=field,
        title=title,
        width=width,
        formatter=StringFormatter(**strfmt_in)
    ))

columns_changed = []
fields = ["SSSCC", "SAMPNO", "diff", "flag_old", "flag_new", "Comments"]
titles = ["SSSCC", "Bottle", "Residual", "Old", "New", "Comments"]
widths = [40, 20, 40, 20, 20, 200]
for (field, title, width) in zip(fields, titles, widths):
    if field == "flag_old":
        strfmt_in = {"text_align": "center", "font_style": "bold"}
    elif field == "flag_new":
        strfmt_in = {"text_align": "center", "font_style": "bold", "text_color": "red"}
    elif field == "Comments":
        strfmt_in = {}
    else:
        strfmt_in = {"text_align": "right"}
    columns_changed.append(TableColumn(
        field=field,
        title=title,
        width=width,
        formatter=StringFormatter(**strfmt_in)
    ))

data_table = DataTable(
    source=src_table,
    columns=columns,
    index_width=20,
    width=480 + 20,  # sum of col widths + idx width
    height=600,
    editable=True,
    fit_columns=True,
    sortable=False,
)
data_table_changed = DataTable(
    source=src_table_changes,
    columns=columns_changed,
    index_width=20,
    width=480 + 20,  # sum of col widths + idx width
    height=200,
    editable=False,
    fit_columns=True,
    sortable=False,
)
data_table_title = Div(text="""<b>All Station Data:</b>""", width=200, height=15)
data_table_changed_title = Div(text="""<b>Flagged Data:</b>""", width=200, height=15)

controls = column(
    parameter,
    ref_param,
    station,
    flag_list,
    bulk_flag_text,
    flag_input,
    flag_button,
    comment_box,
    comment_button,
    vspace,
    save_button,
    width=170,
)
tables = column(
    data_table_title, data_table, data_table_changed_title, data_table_changed
)

curdoc().add_root(row(controls, tables, fig))
curdoc().title = "CTDO Data Flagging Tool"

update_selectors()



import logging
from pathlib import Path

import cmocean
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd

log = logging.getLogger(__name__)

#####
# The following section is a little brittle due to hardcoded names, but we'll fix
# that later. Code copy pasted from jupyter notebook.
#####


def make_cruise_report_plots(df):
    """Create and output all plots"""
    btl_t1_residuals_pressure_plot(df)
    btl_t2_residuals_pressure_plot(df)
    t1_t2_residuals_pressure_plot(df)
    btl_t1_residuals_station_plot(df)
    btl_t2_residuals_station_plot(df)
    t1_t2_residuals_station_plot(df)
    btl_t1_residuals_station_deep_plot(df)
    btl_t2_residuals_station_deep_plot(df)
    t1_t2_residuals_station_deep_plot(df)
    btl_c1_residuals_pressure_plot(df)
    btl_c2_residuals_pressure_plot(df)
    c1_c2_residuals_pressure_plot(df)
    btl_c1_residuals_station_plot(df)
    btl_c2_residuals_station_plot(df)
    c1_c2_residuals_station_plot(df)
    btl_c1_residuals_station_deep_plot(df)
    btl_c2_residuals_station_deep_plot(df)
    c1_c2_residuals_station_deep_plot(df)
    c_t_coherence_plot(df)
    btl_c1_residuals_compare_plot(df)
    btl_c2_residuals_compare_plot(df)
    c1_c2_residuals_compare_plot(df)
    btl_c1_residuals_station_uncorrected_plot(df)
    btl_c2_residuals_station_uncorrected_plot(df)
    c1_c2_residuals_station_uncorrected_plot(df)
    btl_sal_pressure_plot(df)
    btl_sal_station_plot(df)
    btl_sal_station_deep_plot(df)
    btl_oxy_residuals_pressure_plot(df)
    btl_oxy_residuals_station_plot(df)
    btl_oxy_residuals_station_deep_plot(df)
    btl_oxy_residuals_temperature_plot(df)
    btl_oxy_residuals_station_temperature_plot(df)
    btl_oxy_residuals_station_deep_temperature_plot(df)
    btl_oxy_residuals_pressure_concentration_plot(df)
    btl_oxy_residuals_station_concentration_plot(df)
    return None

    #################################################################
    ##### Here lies the temperature plots, long may they rest.  #####
    #################################################################


def btl_t1_residuals_pressure_plot(reft_vals, t1_vals, press, stnno):

    reft_t1 = reft_vals - t1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(reft_t1, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("REFTMP-CTDTMP1 vs CTDPRS")
    ax.set_xlabel("T1 Residual (T90 C)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/reftmp_t1_p.svg", format="svg")
    fig.savefig("./data/images/reftmp_t1_p.pdf", format="pdf")
    plt.close()
    return None


def btl_t2_residuals_pressure_plot(reft_vals, t2_vals, press, stnno):

    reft_t2 = reft_vals - t2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(reft_t2, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("REFTMP-CTDTMP2 vs CTDPRS")
    ax.set_xlabel("T2 Residual (T90 C)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/reftmp_t2_p.svg", format="svg")
    fig.savefig("./data/images/reftmp_t2_p.pdf", format="pdf")
    plt.close()
    return None


def t1_t2_residuals_pressure_plot(t1_vals, t2_vals, press, stnno):

    t1_t2 = t1_vals - t2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(t1_t2, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("CTDTMP1-CTDTMP2 vs CTDPRS")
    ax.set_xlabel("T1-T2 Residual (T90 C)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/t1_t2_p.svg", format="svg")
    fig.savefig("./data/images/t1_t2_p.pdf", format="pdf")
    plt.close()
    return None


def btl_t1_residuals_station_plot(reft_vals, t1_vals, press, stnno):

    reft_t1 = reft_vals - t1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, reft_t1, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("REFTMP-CTDTMP1 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T1 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/reftmp_t1_stn.svg", format="svg")
    fig.savefig("./data/images/reftmp_t1_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_t2_residuals_station_plot(t1_vals, t2_vals, press, stnno):

    t1_t2 = t1_vals - t2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, t1_t2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("REFTMP-CTDTMP2 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T2 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/reftmp_t2_stn.svg", format="svg")
    fig.savefig("./data/images/reftmp_t2_stn.pdf", format="pdf")
    plt.close()
    return None


def t1_t2_residuals_station_plot(t1_vals, t2_vals, press, stnno):

    t1_t2 = t1_vals - t2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, t1_t2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("CTDTMP1-CTDTMP2 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T1-T2 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/t1_t2_stn.svg", format="svg")
    fig.savefig("./data/images/t1_t2_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_t1_residuals_station_deep_plot(reft_vals, t1_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["REFT_T1"] = reft_vals - t1_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["REFT_T1"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("REFTMP-CTDTMP1 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T1 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/reftmp_t1_stn_deep.svg", format="svg")
    fig.savefig("./data/images/reftmp_t1_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def btl_t2_residuals_station_deep_plot(reft_vals, t2_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["REFT_T2"] = reft_vals - t2_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["REFT_T2"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("REFTMP-CTDTMP2 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T2 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/reftmp_t2_stn_deep.svg", format="svg")
    fig.savefig("./data/images/reftmp_t2_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def t1_t2_residuals_station_deep_plot(t1_vals, t2_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["T1_T2"] = t1_vals - t2_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["T1_T2"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("CTDTMP1-CTDTMP2 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("T1-T2 Residual (T90 C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/t1_t2_stn_deep.svg", format="svg")
    fig.savefig("./data/images/t1_t2_stn_deep.pdf", format="pdf")
    plt.close()
    return None

    #################################################################
    ##### Here lies the conductivity plots, long may they rest. #####
    #################################################################


def btl_c1_residuals_pressure_plot(refc_vals, c1_vals, press, stnno):

    refc_c1 = refc_vals - c1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(refc_c1, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("BTLCOND-CTDCOND1 vs CTDPRS")
    ax.set_xlabel("C1 Residual (mS/cm)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btlcond_c1_p.svg", format="svg")
    fig.savefig("./data/images/btlcond_c1_p.pdf", format="pdf")
    plt.close()
    return None


def btl_c2_residuals_pressure_plot(refc_vals, c2_vals, press, stnno):

    refc_c2 = refc_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(refc_c2, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_title("BTLCOND-CTDCOND2 vs CTDPRS")
    ax.set_xlabel("C2 Residual (mS/cm)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btlcond_c2_p.svg", format="svg")
    fig.savefig("./data/images/btlcond_c2_p.pdf", format="pdf")
    plt.close()
    return None


def c1_c2_residuals_pressure_plot(c1_vals, c2_vals, press, stnno):

    c1_c2 = c1_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(c1_c2, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("CTDCOND1-CTDCOND2 vs CTDPRS")
    ax.set_xlabel("C1-C2 Residual (mS/cm)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/c1_c2_p.svg", format="svg")
    fig.savefig("./data/images/c1_c2_p.pdf", format="pdf")
    plt.close()
    return None


def btl_c1_residuals_station_plot(refc_vals, c1_vals, press, stnno):

    refc_c1 = refc_vals - c1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, refc_c1, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND1 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c1_stn.svg", format="svg")
    fig.savefig("./data/images/btlcond_c1_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_c2_residuals_station_plot(refc_vals, c2_vals, press, stnno):

    refc_c2 = refc_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, refc_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND2 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c2_stn.svg", format="svg")
    fig.savefig("./data/images/btlcond_c2_stn.pdf", format="pdf")
    plt.close()
    return None


def c1_c2_residuals_station_plot(c1_vals, c2_vals, press, stnno):

    c1_c2 = c1_vals - c2_vals
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, c1_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("CTDCOND1-CTDCOND2 vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1-C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/c1_c2_stn.svg", format="svg")
    fig.savefig("./data/images/c1_c2_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_c1_residuals_station_deep_plot(refc_vals, c1_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["REFT_C1"] = refc_vals - c1_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["REFT_C1"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND1 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c1_stn_deep.svg", format="svg")
    fig.savefig("./data/images/btlcond_c1_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def btl_c2_residuals_station_deep_plot(refc_vals, c2_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["REFT_C2"] = refc_vals - c2_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["REFT_C2"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND2 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c2_stn_deep.svg", format="svg")
    fig.savefig("./data/images/btlcond_c2_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def c1_c2_residuals_station_deep_plot(c1_vals, c2_vals, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["C1_C2"] = c1_vals - c2_vals
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["C1_C2"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("CTDCOND1-CTDCOND2 (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1-C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/c1_c2_stn_deep.svg", format="svg")
    fig.savefig("./data/images/c1_c2_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def c_t_coherence_plot(t1_vals, t2_vals, c1_vals, c2_vals, press):

    t1_t2 = t1_vals - t2_vals
    c1_c2 = c1_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(t1_t2, c1_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("T1-T2 vs C1-C2")
    ax.set_xlabel("T1-T2 Residual (T90 C)")
    ax.set_ylabel("C1-C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/c_t_coherence_p.svg", format="svg")
    fig.savefig("./data/images/c_t_coherence_p.pdf", format="pdf")
    plt.close()
    return None


def btl_c1_residuals_compare_plot(refc_vals, c1_vals, press):

    refc_c1 = refc_vals - c1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(refc_vals, refc_c1, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND vs BTLCOND-CTDCOND1")
    ax.set_xlabel("Reference Conductivity (mS/cm)")
    ax.set_ylabel("C1 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c1_compare.svg", format="svg")
    fig.savefig("./data/images/btlcond_c1_compare.pdf", format="pdf")
    plt.close()
    return None


def btl_c2_residuals_compare_plot(refc_vals, c2_vals, press):

    refc_c2 = refc_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(refc_vals, refc_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND vs BTLCOND-CTDCOND2")
    ax.set_xlabel("Reference Conductivity (mS/cm)")
    ax.set_ylabel("C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c2_compare.svg", format="svg")
    fig.savefig("./data/images/btlcond_c2_compare.pdf", format="pdf")
    plt.close()
    return None


def c1_c2_residuals_compare_plot(refc_vals, c1_vals, c2_vals, press):

    c1_c2 = c1_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(refc_vals, c1_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND vs CTDCOND1-CTDCOND2")
    ax.set_xlabel("Reference Conductivity (mS/cm)")
    ax.set_ylabel("C1-C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/c1_c2_compare.svg", format="svg")
    fig.savefig("./data/images/c1_c2_compare.pdf", format="pdf")
    plt.close()
    return None


def btl_c1_residuals_station_uncorrected_plot(refc_vals, c1_vals, press, stnno):

    refc_c1 = refc_vals - c1_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, refc_c1, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND1 (Uncorrected) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c1_stn_uncorrected.svg", format="svg")
    fig.savefig("./data/images/btlcond_c1_stn_uncorrected.pdf", format="pdf")
    plt.close()
    return None


def btl_c2_residuals_station_uncorrected_plot(refc_vals, c2_vals, press, stnno):

    refc_c2 = refc_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, refc_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("BTLCOND-CTDCOND2 (Uncorrected) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlcond_c2_stn_uncorrected.svg", format="svg")
    fig.savefig("./data/images/btlcond_c2_stn_uncorrected.pdf", format="pdf")
    plt.close()
    return None


def c1_c2_residuals_station_uncorrected_plot(c1_vals, c2_vals, press, stnno):

    c1_c2 = c1_vals - c2_vals

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, c1_c2, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("CTDCOND1-CTDCOND2 (Uncorrected) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("C1-C2 Residual (mS/cm)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/c1_c2_stn_uncorrected.svg", format="svg")
    fig.savefig("./data/images/c1_c2_stn_uncorrected.pdf", format="pdf")
    plt.close()
    return None


def btl_sal_pressure_plot(btl_sal, ctd_sal, press, stnno):
    sal_res = btl_sal - ctd_sal
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(sal_res, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-0.02, 0.02)
    ax.set_title("SALNTY-CTDSAL vs CTDPRS")
    ax.set_xlabel("CTDSAL Residual (mPSU)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btlsal_sal_p.svg", format="svg")
    fig.savefig("./data/images/btlsal_sal_p.pdf", format="pdf")
    plt.close()
    return None


def btl_sal_station_plot(btl_sal, ctd_sal, press, stnno):
    sal_res = btl_sal - ctd_sal
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, sal_res, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("SALNTY-CTDSAL vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDSAL Residual (mPSU)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlsal_sal_stn.svg", format="svg")
    fig.savefig("./data/images/btlsal_sal_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_sal_station_deep_plot(btl_sal, ctd_sal, press, stnno):
    sal_res = btl_sal - ctd_sal

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["BTL_SAL"] = sal_res
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["BTL_SAL"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-0.01, 0.01)
    ax.set_title("SALNTY-CTDSAL (>2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDSAL Residual (mPSU)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btlsal_sal_stn_deep.svg", format="svg")
    fig.savefig("./data/images/btlsal_sal_stn_deep.pdf", format="pdf")
    plt.close()
    return None

    #################################################################
    ######## Here lies the oxygen plots, long may they rest. ########
    #################################################################


def btl_oxy_residuals_pressure_plot(ref_oxy, ctdoxy, press, stnno):

    btl_o = ref_oxy - ctdoxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(btl_o, -press, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs CTDPRS")
    ax.set_xlabel("CTDOXY Residual (umol/kg)")
    ax.set_ylabel("Pressure (dbar)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btl_oxy_p.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_p.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_station_plot(ref_oxy, ctdoxy, press, stnno):

    btl_o = ref_oxy - ctdoxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, btl_o, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_ylim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDOXY Residual (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btl_oxy_stn.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_stn.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_station_deep_plot(ref_oxy, ctdoxy, press, stnno):

    df = pd.DataFrame()
    df["CTDPRS"] = press
    df["BTL_O"] = ref_oxy - ctdoxy
    df["STNNBR"] = stnno

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["BTL_O"],
        marker="+",
        c=df_deep["CTDPRS"],
        cmap=plt.cm.viridis_r,
    )
    ax.set_ylim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY (> 2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDOXY Residual (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btl_oxy_stn_deep.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_stn_deep.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_temperature_plot(ref_oxy, ctd_oxy, t1_vals, stnno):

    btl_o = ref_oxy - ctd_oxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(btl_o, t1_vals, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs CTDPRS")
    ax.set_xlabel("CTDOXY Residual (umol/kg)")
    ax.set_ylabel("Temperature (degrees C)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btl_oxy_t.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_t.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_station_temperature_plot(ref_oxy, ctd_oxy, t1_vals, stnno):

    btl_o = ref_oxy - ctd_oxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(stnno, btl_o, marker="+", c=t1_vals, cmap=cmocean.cm.thermal)
    ax.set_ylim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDOXY Residual (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Temperature (degrees C)")

    fig.savefig("./data/images/btl_oxy_stn_t.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_stn_t.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_station_deep_temperature_plot(
    ref_oxy, ctdoxy, t1_vals, press, stnno
):

    df = pd.DataFrame()
    df["CTDTMP1"] = t1_vals
    df["BTL_O"] = ref_oxy - ctdoxy
    df["STNNBR"] = stnno
    df["CTDPRS"] = press

    df_deep = df[df["CTDPRS"] > 2000]
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(
        df_deep["STNNBR"],
        df_deep["BTL_O"],
        marker="+",
        c=df_deep["CTDTMP1"],
        cmap=cmocean.cm.thermal,
    )
    ax.set_ylim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY (> 2000 db) vs STNNBR")
    ax.set_xlabel("Station Number")
    ax.set_ylabel("CTDOXY Residual (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Temperature (degrees C)")

    fig.savefig("./data/images/btl_oxy_stn_deep_t.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_stn_deep_t.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_pressure_concentration_plot(ref_oxy, ctd_oxy, stnno):

    btl_o = ref_oxy - ctd_oxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(btl_o, ref_oxy, marker="+", c=stnno, cmap=plt.cm.tab20c_r)
    ax.set_xlim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs OXYGEN")
    ax.set_xlabel("CTDOXY Residual (umol/kg)")
    ax.set_ylabel("Dissolved Oxygen (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Station Number")

    fig.savefig("./data/images/btl_oxy_p_concentration.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_p_concentration.pdf", format="pdf")
    plt.close()
    return None


def btl_oxy_residuals_station_concentration_plot(ref_oxy, ctd_oxy, press, stnno):

    btl_o = ref_oxy - ctd_oxy

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    cm = ax.scatter(btl_o, ref_oxy, marker="+", c=press, cmap=plt.cm.viridis_r)
    ax.set_xlim(-10, 10)
    ax.set_title("OXYGEN-CTDOXY vs OXYGEN")
    ax.set_xlabel("CTDOXY Residual (umol/kg)")
    ax.set_ylabel("Dissolved Oxygen (umol/kg)")
    cbar = fig.colorbar(cm)
    cbar.set_label("Pressure (dbar)")

    fig.savefig("./data/images/btl_oxy_stn_concentration.svg", format="svg")
    fig.savefig("./data/images/btl_oxy_stn_concentration.pdf", format="pdf")
    plt.close()
    return None


def residual_vs_pressure(
    param,
    ref,
    prs,
    stn,
    xlim=(-0.02, 0.02),
    ylim=(6000, 0),
    xlabel="Residual",
    ylabel="Pressure (dbar)",
    deep=False,
    f_out=None,
):

    title = f"{ref.name}-{param.name} vs. {prs.name}"
    diff = ref - param
    if deep:
        title = f"{ref.name}-{param.name} (>2000 dbar) vs. {prs.name}"
        deep_rows = prs > 2000
        diff = diff[deep_rows]
        prs = prs[deep_rows]
        stn = stn[deep_rows]

    idx, uniques = stn.factorize()  # find unique stations #s and index them

    plt.figure(figsize=(7, 6))
    plt.scatter(diff, prs, c=idx, marker="+")
    plt.xlim(xlim)
    plt.xticks(rotation=45)
    plt.ylim(ylim)
    cbar = plt.colorbar(pad=0.1)  # set cbar ticks to station names
    if not uniques.empty:
        tick_inds = cbar.get_ticks().astype(int)
        cbar.ax.yaxis.set_major_locator(ticker.FixedLocator(tick_inds))
        cbar.ax.set_yticklabels(uniques[tick_inds])
    cbar.set_label("Station Number")
    plt.xlabel(xlabel, fontsize=12)
    plt.ylabel(ylabel, fontsize=12)
    plt.title(title, fontsize=12)
    plt.tight_layout()
    if f_out is not None:
        if not Path(f_out).parent.exists():
            log.info(
                f"Path {Path(f_out).parent.as_posix()} does not exists... creating"
            )
            Path(f_out).parent.mkdir(parents=True)
        plt.savefig(f_out)
    plt.close()

    return True


def residual_vs_station(
    param,
    ref,
    prs,
    stn,
    ylim=(-0.02, 0.02),
    xlabel="Station Number",
    ylabel="Residual",
    deep=False,
    f_out=None,
):

    title = f"{ref.name}-{param.name} vs. {stn.name}"
    diff = ref - param
    if deep:
        title = f"{ref.name}-{param.name} (>2000 dbar) vs. {stn.name}"
        deep_rows = prs > 2000
        diff = diff[deep_rows]
        prs = prs[deep_rows]
        stn = stn[deep_rows]

    plt.figure(figsize=(7, 6))
    plt.scatter(stn, diff, c=prs, marker="+")
    plt.xticks(rotation=45)
    plt.ylim(ylim)
    cbar = plt.colorbar(pad=0.1)
    cbar.set_label("Pressure (dbar)")
    plt.xlabel(xlabel, fontsize=12)
    plt.ylabel(ylabel, fontsize=12)
    plt.title(title, fontsize=12)
    plt.tight_layout()
    if f_out is not None:
        if not Path(f_out).parent.exists():
            log.info(
                f"Path {Path(f_out).parent.as_posix()} does not exists... creating"
            )
            Path(f_out).parent.mkdir(parents=True)
        plt.savefig(f_out)
    plt.close()

    return True


def _intermediate_residual_plot(
    diff,
    prs,
    ssscc,
    xlim=(-0.02, 0.02),
    ylim=(6000, 0),
    xlabel="Residual",
    ylabel="CTDPRS",
    show_thresh=False,
    f_out=None,
):

    idx, uniques = ssscc.factorize()  # find unique SSSCC and index them

    plt.figure(figsize=(6, 6))
    plt.scatter(diff, prs, c=idx, marker="+")
    if show_thresh:
        # TODO: thresh should probably be put in config/cast-by-cast config
        thresh = np.array([0.002, 0.005, 0.010, 0.020])
        p_range = np.array([6000, 2000, 1000, 500])
        thresh = np.append(thresh, thresh[-1])  # this should still work fine even when
        p_range = np.append(p_range, 0)  # thresh/p_range are defined elsewhere
        plt.step(thresh, p_range, ":k")
        plt.step(-thresh, p_range, ":k")

    plt.xlim(xlim)
    plt.xticks(rotation=45)
    plt.ylim(ylim)
    cbar = plt.colorbar(pad=0.1)  # set cbar ticks to SSSCC names
    if not uniques.empty:
        tick_inds = cbar.get_ticks().astype(int)
        cbar.ax.yaxis.set_major_locator(ticker.FixedLocator(tick_inds))
        cbar.ax.set_yticklabels(uniques[tick_inds])
        mean = np.round(np.nanmean(diff), 4)
        stdev = np.round(np.nanstd(diff), 4)
        plt.title(f"Mean: {mean} / Stdev: {stdev}")
    cbar.ax.set_title("SSSCC")
    plt.grid()
    plt.xlabel(xlabel, fontsize=12)
    plt.ylabel(ylabel, fontsize=12)
    plt.tight_layout()
    if f_out is not None:
        if not Path(f_out).parent.exists():
            log.info(
                f"Path {Path(f_out).parent.as_posix()} does not exists... creating"
            )
            Path(f_out).parent.mkdir(parents=True)
        plt.savefig(f_out)
    plt.close()

    return True



import csv
import datetime as dt
import logging
import sys
from collections import OrderedDict
from pathlib import Path

import gsw
import numpy as np
import pandas as pd

from . import get_ctdcal_config

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)


def _salt_loader(ssscc, salt_dir):
    """
    Load raw file into salt and reference DataFrames.
    """
    saltpath = salt_dir + ssscc  # salt files have no extension
    with open(saltpath, newline="") as f:
        saltF = csv.reader(
            f, delimiter=" ", quoting=csv.QUOTE_NONE, skipinitialspace="True"
        )
        saltArray = []
        for row in saltF:
            saltArray.append(row)
        del saltArray[0]  # remove header

    header = OrderedDict(  # having this as a dict streamlines next steps
        [
            ("STNNBR", int),
            ("CASTNO", int),
            ("SAMPNO", int),
            ("BathTEMP", int),
            ("CRavg", float),
            ("autosalSAMPNO", int),
            ("Unknown", int),
            ("StartTime", object),
            ("EndTime", object),
            ("Attempts", int),
        ]
    )
    saltDF = pd.DataFrame.from_records(saltArray)

    # add as many "Reading#"s as needed
    for ii in range(0, len(saltDF.columns) - len(header)):
        header["Reading{}".format(ii + 1)] = float
    saltDF.columns = list(header.keys())  # name columns

    # TODO: check autosalSAMPNO against SAMPNO for mismatches?
    # TODO: handling for re-samples?

    # check for commented out lines
    commented = saltDF["STNNBR"].str.startswith(("#", "x"))
    if commented.any():
        log.debug(f"Found comment character (#, x) in {ssscc} salt file, ignoring line")
        saltDF = saltDF[~commented]

    # check end time for * and code questionable
    # (unconfirmed but * appears to indicate a lot of things from LabView code:
    # large spread in values, long time between samples, manual override, etc.)
    flagged = saltDF["EndTime"].str.contains("*", regex=False)
    if flagged.any():
        # remove asterisks from EndTime and flag samples
        log.debug(f"Found * in {ssscc} salt file, flagging value(s) as questionable")
        saltDF["EndTime"] = saltDF["EndTime"].str.rstrip("*")
        questionable = pd.DataFrame()
        questionable["SAMPNO"] = saltDF.loc[flagged, "SAMPNO"].astype(int)
        questionable.insert(0, "SSSCC", ssscc)
        questionable["diff"] = np.nan
        questionable["salinity_flag"] = 3
        questionable["comments"] = "Auto-flagged by processing function (had * in row)"
        questionable.to_csv(
            "tools/salt_flags_handcoded.csv", mode="a+", index=False, header=None
        )

    # add time (in seconds) needed for autosal drift removal step
    saltDF["IndexTime"] = pd.to_datetime(saltDF["EndTime"])
    saltDF["IndexTime"] = (saltDF["IndexTime"] - saltDF["IndexTime"].iloc[0]).dt.seconds
    saltDF["IndexTime"] += (saltDF["IndexTime"] < 0) * (3600 * 24)  # fix overnight runs

    refDF = saltDF.loc[
        saltDF["autosalSAMPNO"] == "worm", ["IndexTime", "CRavg"]
    ].astype(float)
    saltDF = saltDF[saltDF["autosalSAMPNO"] != "worm"].astype(header)  # force dtypes

    return saltDF, refDF


def remove_autosal_drift(saltDF, refDF):
    """Calculate linear CR drift between reference values"""
    if len(refDF) != 2:
        ssscc = f"{saltDF['STNNBR'].unique()[0]:03d}{saltDF['CASTNO'].unique()[0]:02d}"
        log.warning(
            f"Failed to find start/end reference readings for {ssscc}, check salt file"
        )

        return saltDF.drop(labels="IndexTime", axis="columns")

    diff = refDF.diff(axis="index").dropna()
    time_coef = (diff["CRavg"] / diff["IndexTime"]).iloc[0]

    saltDF["CRavg"] += saltDF["IndexTime"] * time_coef
    saltDF["CRavg"] = saltDF["CRavg"].round(5)  # match initial precision
    saltDF = saltDF.drop(labels="IndexTime", axis="columns")

    return saltDF


def _salt_exporter(
    saltDF, outdir=cfg.dirs["salt"], stn_col="STNNBR", cast_col="CASTNO"
):
    """
    Export salt DataFrame to .csv file. Extra logic is included in the event that
    multiple stations and/or casts are included in a single raw salt file.
    """
    stations = saltDF[stn_col].unique()
    for station in stations:
        stn_salts = saltDF[saltDF[stn_col] == station]
        casts = stn_salts[cast_col].unique()
        for cast in casts:
            stn_cast_salts = stn_salts[stn_salts[cast_col] == cast]
            stn_cast_salts.dropna(axis=1, how="all", inplace=True)  # drop empty columns
            outfile = (  # format to SSSCC_salts.csv
                outdir + "{0:03}".format(station) + "{0:02}".format(cast) + "_salts.csv"
            )
            if Path(outfile).exists():
                log.debug(outfile + " already exists...skipping")
                continue
            stn_cast_salts.to_csv(outfile, index=False)


def process_salts(ssscc_list, salt_dir=cfg.dirs["salt"]):
    """
    Master salt processing function. Load in salt files for given station/cast list,
    calculate salinity, and export to .csv files.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to process
    salt_dir : str, optional
        Path to folder containing raw salt files (defaults to data/salt/)

    """
    for ssscc in ssscc_list:
        if not Path(salt_dir + ssscc + "_salts.csv").exists():
            try:
                saltDF, refDF = _salt_loader(ssscc, salt_dir)
            except FileNotFoundError:
                log.warning(f"Salt file for cast {ssscc} does not exist... skipping")
                continue
            saltDF = remove_autosal_drift(saltDF, refDF)
            saltDF["SALNTY"] = gsw.SP_salinometer(
                (saltDF["CRavg"] / 2.0), saltDF["BathTEMP"]
            )#.round(4)
            _salt_exporter(saltDF, salt_dir)

    return True



import pandas as pd
import logging

logging.basicConfig(filename='errors/compare_summary.txt', format='%(asctime)s:%(levelname)s:%(message)s', level=logging.INFO)
#0 for diffs of columns that stand out, 1 for full summaries and diffs
DEBUG_FLAG = 1
ssscc = ['00101', '00102', '00201', '00202', '00301', '00401', '00402', '00501', '00601', '00602', '00701', '00801', '00901', '01001', '01101', '01201', '01301', '01401', '01402', '01501', '01502', '01601','01602','01701','01702','01801','01802','01901','01904','02001','02002','02101','02201','02301','02401','02402','02501','02601','02602','02801','02901','03001','03101','03201']

def compare_all(ssscc, sequence='pressure', verbosity=0):
    for x in ssscc:
        print('\nCast number: ' + x)
        #Start with time first, then pressure
        if sequence == 'time':
            df1 = pd.read_csv('data/time/' + x + '_time.csv', header = [1,2])
            df2 = pd.read_csv('data/initial/time/' + x + '_time.csv', header = [1,2])
        if sequence == 'pressure':
            df1 = pd.read_csv('data/pressure/' + x + '.csv', header = [12,13])
            df2 = pd.read_csv('data/initial/pressure/' + x + '.csv', header = [12,13])

        #Find the difference of the values.
        #TODO - try to determine in df3 where the biggest discrepancies happened
        try:
            df3 = df1.subtract(df2)
            series_sum = df3.sum()

            #show results to user
            if verbosity >= 1:
                print(series_sum)

            if verbosity >= 0:
                #if the difference of values is greater than 0.0, throw it up for checking
                unchanged_file_counter = 0
                for i, y in enumerate(series_sum):
                    if y != 0.0:
                        print(str(series_sum.index[i]) + ' difference: ' + str(y))
                        unchanged_file_counter += 1
                if unchanged_file_counter == 0:
                    logging.info('Cast {} {} is unchanged after fitting'.format(x,sequence))
        except TypeError:
            print('Not all data points are same type, skip for now')

if __name__ == "__main__":
    compare_all(ssscc, 'time', DEBUG_FLAG)
    print('Start pressure differences')
    compare_all(ssscc,'pressure', DEBUG_FLAG)



import sys

import numpy as np
import pandas as pd


def main(argv):
    """Creates a bottle file with CTD downcast information.

    Needs to be cleaned up from script form to other form, hopefully
    """
    # general fileloading area
    output_file = "data/scratch_folder/ctd_to_bottle.csv"
    file_ssscc = "data/ssscc.csv"
    df_all = pd.DataFrame()

    all_casts = []
    with open(file_ssscc, "r") as filename:
        all_casts = [line.strip() for line in filename]

    for ssscc in all_casts:
        station = int(ssscc[0:3])
        cast = int(ssscc[3:5])
        # bottle handling section
        dir_bottle = "data/bottle/"
        bottle_postfix = "_btl_mean.pkl"

        df_bottle = pd.read_pickle(f"{dir_bottle}{ssscc}{bottle_postfix}")
        # next line not strictly necessaryas we don't move every column, but left just in case
        df_bottle.rename(
            index=str,
            columns={
                "FREE1": "CTDFLUOR",
                "FREE2": "CTDBACKSCATTER",
                "FREE3": "CTDRINKO",
                "btl_fire_num": "SAMPNO",
                "CTDTMP1": "CTDTMP",
                "CTDOXY1": "CTDOXY",
            },
            inplace=True,
        )
        df_bottle = df_bottle.loc[:, ["CTDPRS", "SAMPNO"]]
        df_bottle["bins"] = pd.cut(
            df_bottle["CTDPRS"],
            range(0, int(np.ceil(df_bottle["CTDPRS"].max())) + 5, 2),
            right=False,
            include_lowest=True,
        )

        # ctd handling section
        dir_ctd = "data/pressure/"
        ctd_postfix = "_ct1.csv"
        ctd_skiprows = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13]

        df_c = pd.read_csv(
            dir_ctd + ssscc + ctd_postfix,
            skiprows=ctd_skiprows,
            skipfooter=1,
            engine="python",
        )
        df_c["bins"] = pd.cut(
            df_c["CTDPRS"],
            range(0, int(np.floor(df_c["CTDPRS"].max())) + 5, 2),
            right=False,
            include_lowest=True,
        )

        df_merged_btl_ctd_4 = pd.merge(
            df_c, df_bottle[["bins"]], on="bins", how="outer"
        )

        df_merged_btl_ctd_6 = pd.merge(
            df_merged_btl_ctd_4.iloc[:, 1:], df_bottle, on="bins", how="inner"
        )
        df_merged_btl_ctd_6.groupby("SAMPNO").first()

        # reference temperature section
        # there should not be any 9 values, unless the reftemp dies? then we have a PROBLEM
        dir_reft = "data/reft/"
        reft_postfix = "_reft.csv"
        try:
            df_reft = pd.read_csv(
                dir_reft + ssscc + reft_postfix,
                names=["SAMPNO", "REFTMP"],
                skiprows=1,
                usecols=[2, 5],
            )
            df_reft["REFTMP_FLAG_W"] = 2
            df_incomplete_reft = pd.merge(
                df_reft, df_merged_btl_ctd_6, on="SAMPNO", how="outer"
            )
            # this line breaks code; seem redundant? MK
            # df_incomplete_reft.groupby('SAMPNO').fillna(inplace=True, method='pad')
            df_complete_reft = df_incomplete_reft.fillna(
                value={"REFTMP": -999, "REFTMP_FLAG_W": 5}
            )
        # if there are no SBE35 files for the cast, flag as 5
        except FileNotFoundError:
            df_merged_btl_ctd_6["REFTMP"] = -999
            df_merged_btl_ctd_6["REFTMP_FLAG_W"] = 5
            df_complete_reft = df_merged_btl_ctd_6

        df_complete_reft["STNNBR"] = station
        df_complete_reft["CASTNO"] = cast
        df_complete_reft = df_complete_reft.groupby("SAMPNO").first()
        df_complete_reft.reset_index(inplace=True)
        df_complete_reft = df_complete_reft.sort_index(axis=1)
        df_all = df_all.append(df_complete_reft)
        # turn any nans to flag 9 values
        fillna_values = {
            "CTDBACKSCATTER": -999,
            "CTDBACKSCATTER_FLAG_W": 9,
            "CTDFLUOR": -999,
            "CTDFLUOR_FLAG_W": 9,
            "CTDOXY": -999,
            "CTDOXY_FLAG_W": 9,
            "CTDPRS": -999,
            "CTDPRS_FLAG_W": 9,
            "CTDRINKO": -999,
            "CTDRINKO_FLAG_W": 9,
            "CTDSAL": -999,
            "CTDSAL_FLAG_W": 9,
            "CTDTMP": -999,
            "CTDTMP_FLAG_W": 9,
            "CTDXMISS": -999,
            "CTDXMISS_FLAG_W": 9,
            "REFTMP": -999,
            "REFTMP_FLAG_W": 9,
        }
        df_all = df_all.fillna(fillna_values)

    # create the stupid string
    # stringy = "CASTNO,CTDBACKSCATTER,CTDBACKSCATTER_FLAG_W,CTDFLUOR,CTDFLUOR_FLAG_W,"
    # MK 03/27/21: no backscatter on A20/22
    stringy = "CASTNO,CTDFLUOR,CTDFLUOR_FLAG_W,"
    stringy += (
        # "CTDOXY,CTDOXY_FLAG_W,CTDPRS,CTDRINKO,CTDRINKO_FLAG_W,CTDSAL,"
        "CTDOXY,CTDOXY_FLAG_W,CTDPRS,CTDSAL,"
    )
    stringy += "CTDSAL_FLAG_W,CTDTMP,CTDTMP_FLAG_W,CTDXMISS,CTDXMISS_FLAG_W,REFTMP,REFTMP_FLAG_W,"
    # stringy += "SAMPNO,STNNBR\n,VOLTS,,VOLTS,,UMOL/KG,,DBAR,,VOLTS,,PSS-78,,ITS-90,,VOLTS,,ITS-90,,,\n"
    stringy += "SAMPNO,STNNBR\n,VOLTS,,UMOL/KG,,DBAR,,VOLTS,,PSS-78,,ITS-90,,VOLTS,,ITS-90,,,\n"
    # the next line is incredibly stupid. find a better way to create
    stringy += df_all.iloc[:, 0:-1].to_string(
        header=False,
        index=False,
        formatters=[
            # lambda x: str(x) + ",",
            # lambda x: str(x) + ",",
            # lambda x: str(x) + ",",
            # lambda x: str(x) + ",",
            # lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x),
        ],
    )
    stringy += "\nEND_DATA\n"

    with open(output_file, "w") as f_handle:
        f_handle.write(stringy)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
Module for processing oxygen from CTD and bottle samples.
"""

import csv
import logging
import xml.etree.cElementTree as ET
from collections import OrderedDict
from pathlib import Path

import gsw
import numpy as np
import pandas as pd
import scipy

from . import ctd_plots as ctd_plots
from . import flagging as flagging
from . import get_ctdcal_config
from . import process_ctd as process_ctd
from . import sbe_reader as sbe_rd

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)


def load_winkler_oxy(oxy_file):
    """
    Load Winkler oxygen titration data file.

    Parameters
    ----------
    oxy_file : str or Path
        Path to oxygen file

    Returns
    -------
    df : DataFrame
        Oxygen data
    params : list of str
        List of oxygen parameters used in titration
    """

    with open(oxy_file, newline="") as f:
        oxyF = csv.reader(
            f, delimiter=" ", quoting=csv.QUOTE_NONE, skipinitialspace="True"
        )
        oxy_array = []
        for row in oxyF:
            if len(row) > 9:
                row = row[:9]
            oxy_array.append(row)

    # TODO turn params into a dict with useful labels
    params = oxy_array.pop(0)  # save file header info for later (Winkler values)
    cols = OrderedDict(
        [
            ("STNNO_OXY", int),
            ("CASTNO_OXY", int),
            ("BOTTLENO_OXY", int),
            ("FLASKNO", int),
            ("TITR_VOL", float),
            ("TITR_TEMP", float),
            ("DRAW_TEMP", float),
            ("TITR_TIME", int),
            ("END_VOLTS", float),
        ]
    )

    df = pd.DataFrame(oxy_array, columns=cols.keys()).astype(cols)
    df = df[df["BOTTLENO_OXY"] != 99]  # remove "Dummy Data"
    df = df[df["TITR_VOL"] > 0]  # remove "ABORTED DATA"
    df = df.sort_values("BOTTLENO_OXY").reset_index(drop=True)
    df["FLASKNO"] = df["FLASKNO"].astype(str)

    return df, params


def load_flasks(flask_file=cfg.dirs["oxygen"] + "o2flasks.vol", comment="#"):
    """
    Load oxygen flask information from .vol file.

    Parameters
    ----------
    flask_file : str or Path, optional
        Path to flask file
    comment : str, optional
        Identifier signifying line is a comment and should be skipped

    Returns
    -------
    flasks : DataFrame
        Flask numbers and volumes
    """
    with open(flask_file, "r") as f:
        flasks = []
        for line in f:
            is_comment = line.strip().startswith(comment)
            if ("Volume" in line) or is_comment:
                continue
            num, vol = line.strip().split()[:2]  # only need first two cols (#, volume)
            flasks.append([str(num), float(vol)])

    flasks = pd.DataFrame(flasks, columns=["FLASKNO", "FLASK_VOL"])

    return flasks


def correct_flask_vol(flask_vol, t=20.0, glass="borosilicate"):
    """
    Correct flask volume for changes from thermal expansion of glass.

    Parameters
    ----------
    flask_vol : array-like
        Flask volumes at standard temperature (20C)
    t : float, optional
        New temperature to calculate volume
    glass : str, optional
        Type of glass ("borosilicate" or "soft)

    Returns
    -------
    corrected_vol : array-like
        Flask volumes are new temperature

    Notes
    -----
    Flask volume equation from 2007 Best Practices for Ocean CO2 Measurements,
    SOP 13 - Gravimetric calibration of volume contained using water
    """
    alpha = {  # thermal expansion coefficient
        "borosilicate": 1.0e-5,
        "soft": 2.5e-3,
    }
    if glass not in alpha.keys():
        raise KeyError(f"Glass type not found, must be one of {list(alpha.keys())}")
    standard_t = 20.0
    corrected_vol = flask_vol * (1.0 + alpha[glass] * (t - standard_t))

    return corrected_vol


def gather_oxy_params(oxy_file):
    """
    Collect Winkler oxygen measurement parameters from LabVIEW data file headers.

    Parameters
    ----------
    oxy_file : str or Path
        Path to oxygen file

    Returns
    -------
    df : DataFrame
        Oxygen measurement parameters
    """
    with open(oxy_file, newline="") as f:
        header = f.readline()

    param_list = header.split()[:6]
    params = pd.DataFrame(param_list, dtype=float).transpose()
    params.columns = ["V_std", "V_blank", "N_KIO3", "V_KIO3", "T_KIO3", "T_thio"]

    return params


def calculate_bottle_oxygen(ssscc_list, ssscc_col, titr_vol, titr_temp, flask_nums):
    """
    Wrapper function for collecting parameters and calculating oxygen values from
    Winkler titrations.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to process
    ssscc_col : array-like
        Station/cast for each sample taken
    titr_vol : array-like
        Titration volume [mL]
    titr_temp : array-like
        Temperature of titration [degC]
    flask_nums : array-like
        Oxygen flask used for each sample

    Returns
    -------
    oxy_mL_L : array-like
        Oxygen concentration [mL/L]

    Notes
    -----
    Titration equation comes from WHP Operations and Methods, Culberson (1991):
    https://cchdo.github.io/hdo-assets/documentation/manuals/pdf/91_1/culber2.pdf

    """
    params = pd.DataFrame()
    for ssscc in ssscc_list:
        df = gather_oxy_params(cfg.dirs["oxygen"] + ssscc)
        df["SSSCC"] = ssscc
        params = pd.concat([params, df])

    # get flask volumes and merge with titration parameters
    flask_df = load_flasks()  # TODO: volume correction from thermal expansion?
    volumes = pd.merge(flask_nums, flask_df, how="left")["FLASK_VOL"].values
    params = pd.merge(ssscc_col, params, how="left")

    # find 20degC equivalents
    rho_20C = gsw.rho_t_exact(0, 20, 0)
    rho_T_KIO3 = gsw.rho_t_exact(0, params["T_KIO3"], 0)
    N_KIO3_20C = params["N_KIO3"] * (rho_T_KIO3 / rho_20C)

    # TODO: does KIO3 volume get corrected? what is the recorded value?
    # V_KIO3_20C = correct_flask_vol(params["V_KIO3"], t=params["T_KIO3"])

    # calculate O2 concentration (in mL/L)
    E = 5598  # stoichiometric relationship between thio_n and DO
    DO_reg = 0.0017  # correction for oxygen added by reagents
    V_reg = 2.0  # volume of reagents (mL)

    oxy_mL_L = (
        (
            ((titr_vol.values - params["V_blank"]) * params["V_KIO3"] * N_KIO3_20C * E)
            / (params["V_std"] - params["V_blank"])
            - 1000 * DO_reg
        )
    ) / (volumes - V_reg)

    return oxy_mL_L.values


def hysteresis_correction(oxygen, pressure, H1=-0.033, H2=5000, H3=1450, freq=24):
    """
    Remove hysteresis effects from oxygen concentration values.

    Oxygen hysteresis can be corrected before conversion from volts to oxygen
    concentration, see equations_sbe.sbe43_hysteresis_voltage()

    # TODO: should this just be a wrapper that calls sbe43_hysteresis_voltage()?

    Parameters
    ----------
    oxygen : array-like
        Oxygen concentration values
    pressure : array-like
        CTD pressure values (dbar)
    H1 : scalar, optional
        Amplitude of hysteresis correction function (range: -0.02 to -0.05)
    H2 : scalar, optional
        Function constant or curvature function for hysteresis
    H3 : scalar, optional
        Time constant for hysteresis (seconds) (range: 1200 to 2000)
    freq : scalar, optional
        CTD sampling frequency (Hz)

    Returns
    -------
    oxy_corrected : array-like
        Hysteresis-corrected oxygen concentration values (with same units as input)

    Notes
    -----
    See Application Note 64-3 for more information.
    """
    # TODO: vectorize (if possible), will probably require matrix inversion
    dt = 1 / freq
    D = 1 + H1 * (np.exp(pressure / H2) - 1)
    C = np.exp(-1 * dt / H3)

    oxy_corrected = np.zeros(oxygen.shape)
    oxy_corrected[0] = oxygen[0]
    for i in np.arange(1, len(oxygen)):
        oxy_corrected[i] = (
            oxygen[i] + (oxy_corrected[i - 1] * C * D[i]) - (oxygen[i - 1] * C)
        ) / D[i]

    return oxy_corrected


def oxy_ml_to_umolkg(oxy_mL_L, sigma0):
    """Convert dissolved oxygen from units of mL/L to micromol/kg.

    Parameters
    ----------
    oxy_mL_L : array-like
        Dissolved oxygen in units of [mL/L]
    sigma0 : array-like
        Potential density anomaly (i.e. sigma - 1000) referenced to 0 dbar [kg/m^3]

    Returns
    -------
    oxy_umol_kg : array-like
        Dissolved oxygen in units of [umol/kg]

    Notes
    -----
    Conversion value 44660 is exact for oxygen gas and derived from the ideal gas law.
    (c.f. Sea-Bird Application Note 64, pg. 6)
    """

    oxy_umol_kg = oxy_mL_L * 44660 / (sigma0 + 1000)

    return oxy_umol_kg


def oxy_umolkg_to_ml(oxy_umol_kg, sigma0):
    """Convert dissolved oxygen from units of micromol/kg to mL/L.

    Parameters
    ----------
    oxy_umol_kg : array-like
        Dissolved oxygen in units of [umol/kg]
    sigma0 : array-like
        Potential density anomaly (i.e. sigma - 1000) referenced to 0 dbar [kg/m^3]

    Returns
    -------
    oxy_mL_L : array-like
        Dissolved oxygen in units of [mL/L]

    Notes
    -----
    Conversion value 44660 is exact for oxygen gas and derived from the ideal gas law.
    (c.f. Sea-Bird Application Note 64, pg. 6)
    """

    oxy_mL_L = oxy_umol_kg * (sigma0 + 1000) / 44660

    return oxy_mL_L


def calculate_dV_dt(oxy_volts, time, nan_replace=True):
    """
    Calculate the time derivative of oxygen voltage.

    Parameters
    ----------
    oxy_volts : array-like
        Oxygen sensor voltage output
    time : array-like
        Time from oxygen sensor (must be same length as oxy_volts)
    nan_replace : bool, optional
        Replace nans in time derivative with the mean value

    Returns
    -------
    dV_dt : array-like
        Time derivative of oxygen voltage
    """
    # TODO: experiment with dt, filtering
    # Uchida (2008): dV/dt "estimated by linear fits over 2 second intervals"
    # should dt just be 1 / freq? i.e. 1/24 Hz

    dV = np.diff(oxy_volts)  # central differences shorten vectors by 1
    dt = np.diff(time)
    dt[dt == 0] = np.median(dt[dt > 0])  # replace with median to avoid dividing by zero

    dV_dt = dV / dt
    dV_dt = np.insert(dV_dt, 0, 0)  # add zero in front to match original length
    dV_dt[np.isinf(dV_dt)] = np.nan  # this check is probably unnecessary

    if nan_replace:
        dV_dt = np.nan_to_num(dV_dt, nan=np.nanmean(dV_dt))

    # TODO: should we do some kind of filtering? e.g.:
    # (PMEL does this calculation on binned data already so filtering is not the same)
    # a = 1
    # windowsize = 5
    # b = (1 / windowsize) * np.ones(windowsize)
    # filtered_dvdt = scipy.signal.filtfilt(b, a, dv_dt)

    return dV_dt  # filtered_dvdt


def _get_sbe_coef(idx=0):
    """
    Get SBE oxygen coefficients from raw .xmlcon files.
    Defaults to using first station in ssscc.csv file.

    Returns the following tuple of coefficients: Soc, offset, Tau20, Tcor, E
    """
    # TODO: does scipy's minimize function needs a tuple? can this be improved further?

    station = process_ctd.get_ssscc_list()[idx]
    xmlfile = cfg.dirs["raw"] + station + ".XMLCON"

    tree = ET.parse(xmlfile)
    root_eq0 = tree.find(".//CalibrationCoefficients[@equation='0']")  # Owens-Millard
    root_eq1 = tree.find(".//CalibrationCoefficients[@equation='1']")  # SBE equation

    coefs = {c.tag: float(c.text) for c in root_eq1}
    coefs["Tcor"] = float(root_eq0.find("Tcor").text)  # only coef needed from eq0
    keep_keys = ["Soc", "offset", "Tau20", "Tcor", "E"]

    return tuple(coefs[key] for key in keep_keys)


def calculate_weights(pressure):
    """
    Calculate weights (as a function of pressure) for weighted least squares fitting.
    Deep measurements are weighted higher than shallow.

    Parameters
    ----------
    presssure : array-like
        Pressure values of oxygen measurements [dbar]

    Returns
    -------
    weights : array-like
        Weight factor for each pressure value
    """
    # TODO: automatic weight calculation rather than hardcoded (machine learning?)

    epsilon = 1e-5  # small offset to avoid interpolation issues

    # define piecewise weight function dependent on pressure
    p_bins = [
        0,
        100,
        100 + epsilon,
        300,
        300 + epsilon,
        500,
        500 + epsilon,
        1200,
        1200 + epsilon,
        2000,
        2000 + epsilon,
        7000,
    ]
    w_bins = [20, 20, 25, 25, 50, 50, 100, 100, 200, 200, 500, 500]
    wgt = scipy.interpolate.interp1d(p_bins, w_bins)

    weights = wgt(pressure)  # get weights from piecewise function

    return weights


"""code_pruning: should this be here or in equations_sbe? somewhere else?"""


def _PMEL_oxy_eq(coefs, inputs, cc=[1.92634e-4, -4.64803e-2]):
    """
    Modified oxygen equation for SBE 43 used by NOAA/PMEL
    coef[0] = Soc
    coef[1] = Voffset
    coef[2] = Tau20
    coef[3] = Tcorr
    coef[4] = E
    """
    Soc, Voff, Tau20, Tcorr, E = coefs
    oxyvolts, pressure, temp, dvdt, os = inputs
    o2 = (
        Soc
        * (
            oxyvolts
            + Voff
            + Tau20 * np.exp(cc[0] * pressure + cc[1] * (temp - 20)) * dvdt
        )
        * os
        * np.exp(Tcorr * temp)
        * np.exp((E * pressure) / (temp + 273.15))
    )

    return o2


def PMEL_oxy_weighted_residual(coefs, weights, inputs, refoxy, L_norm=2):
    # TODO: optionally include other residual types
    # (abstracted from PMEL code oxygen_cal_ml.m)
    # unweighted L2: sum((ref - oxy)^2)  # if weighted fails
    # unweighted L4: sum((ref - oxy)^4)  # unsure of use case
    # unweighted L1: sum(abs(ref - oxy))  # very far from ideal
    # anything else? genericize with integer "norm" function input?

    residuals = np.sum(
        (weights * (refoxy - _PMEL_oxy_eq(coefs, inputs)) ** 2)
    ) / np.sum(weights ** 2)

    return residuals


def match_sigmas(
    btl_prs,
    btl_oxy,
    btl_tmp,
    btl_SA,
    ctd_os,
    ctd_prs,
    ctd_tmp,
    ctd_SA,
    ctd_oxyvolts,
    ctd_time,
):

    # Construct Dataframe from bottle and ctd values for merging
    btl_data = pd.DataFrame(
        data={"CTDPRS": btl_prs, "REFOXY": btl_oxy, "CTDTMP": btl_tmp, "SA": btl_SA}
    )
    time_data = pd.DataFrame(
        data={
            "CTDPRS": ctd_prs,
            "OS": ctd_os,
            "CTDTMP": ctd_tmp,
            "SA": ctd_SA,
            "CTDOXYVOLTS": ctd_oxyvolts,
            "CTDTIME": ctd_time,
        }
    )
    time_data["dv_dt"] = calculate_dV_dt(time_data["CTDOXYVOLTS"], time_data["CTDTIME"])

    # Merge DF
    merged_df = pd.DataFrame(
        columns=["CTDPRS", "CTDOXYVOLTS", "CTDTMP", "dv_dt", "OS"], dtype=float
    )
    merged_df["REFOXY"] = btl_data["REFOXY"].copy()

    # calculate sigma referenced to multiple depths
    for idx, p_ref in enumerate([0, 1000, 2000, 3000, 4000, 5000, 6000]):

        # pandas 1.2.1 ufunc issue workaround
        btl_inputs = np.broadcast_arrays(
            btl_data["SA"], btl_data["CTDTMP"], btl_data["CTDPRS"], p_ref
        )
        time_inputs = np.broadcast_arrays(
            time_data["SA"], time_data["CTDTMP"], time_data["CTDPRS"], p_ref
        )

        btl_data[f"sigma{idx}"] = (
            gsw.pot_rho_t_exact(*btl_inputs)
            - 1000  # subtract 1000 to get potential density *anomaly*
        ) + 1e-8 * np.random.standard_normal(btl_data["SA"].size)
        time_data[f"sigma{idx}"] = (
            gsw.pot_rho_t_exact(*time_inputs)
            - 1000  # subtract 1000 to get potential density *anomaly*
        ) + 1e-8 * np.random.standard_normal(time_data["SA"].size)
        rows = (btl_data["CTDPRS"] > (p_ref - 500)) & (
            btl_data["CTDPRS"] < (p_ref + 500)
        )

        time_sigma_sorted = time_data[f"sigma{idx}"].sort_values().to_numpy()
        sigma_min = np.min(
            [np.min(btl_data.loc[rows, f"sigma{idx}"]), np.min(time_sigma_sorted)]
        )
        sigma_max = np.max(
            [np.max(btl_data.loc[rows, f"sigma{idx}"]), np.max(time_sigma_sorted)]
        )
        time_sigma_sorted = np.insert(time_sigma_sorted, 0, sigma_min - 1e-4)
        time_sigma_sorted = np.append(time_sigma_sorted, sigma_max + 1e-4)
        # TODO: can this be vectorized?
        cols = ["CTDPRS", "CTDOXYVOLTS", "CTDTMP", "dv_dt", "OS"]
        inds = np.concatenate(([0], np.arange(0, len(time_data)), [len(time_data) - 1]))
        for col in cols:
            merged_df.loc[rows, col] = np.interp(
                btl_data.loc[rows, f"sigma{idx}"],
                time_sigma_sorted,
                time_data[col].iloc[inds],
            )

    # Apply coef and calculate CTDOXY
    sbe_coef0 = _get_sbe_coef()  # initial coefficient guess
    merged_df["CTDOXY"] = _PMEL_oxy_eq(
        sbe_coef0,
        (
            merged_df["CTDOXYVOLTS"],
            merged_df["CTDPRS"],
            merged_df["CTDTMP"],
            merged_df["dv_dt"],
            merged_df["OS"],
        ),
    )

    return merged_df


def sbe43_oxy_fit(merged_df, sbe_coef0=None, f_suffix=None):

    # Plot data to be fit together
    f_out = f"{cfg.fig_dirs['ox']}sbe43_residual{f_suffix}_prefit.pdf"
    ctd_plots._intermediate_residual_plot(
        merged_df["REFOXY"] - merged_df["CTDOXY"],
        merged_df["CTDPRS"],
        merged_df["SSSCC"],
        xlabel="CTDOXY Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )

    bad_df = pd.DataFrame()  # initialize DF for questionable values

    if sbe_coef0 is None:
        sbe_coef0 = _get_sbe_coef()  # load initial coefficient guess

    # Curve fit (weighted)
    weights = calculate_weights(merged_df["CTDPRS"])
    fit_vars = ["CTDOXYVOLTS", "CTDPRS", "CTDTMP", "dv_dt", "OS"]
    fit_data = tuple(merged_df[v] for v in fit_vars)
    res = scipy.optimize.minimize(
        PMEL_oxy_weighted_residual,
        x0=sbe_coef0,
        args=(weights, fit_data, merged_df["REFOXY"]),
        bounds=[(None, None), (None, None), (0, None), (None, None), (None, None)],
    )

    cfw_coefs = res.x
    merged_df["CTDOXY"] = _PMEL_oxy_eq(cfw_coefs, fit_data)
    merged_df["residual"] = merged_df["REFOXY"] - merged_df["CTDOXY"]
    cutoff = 2.8 * np.std(merged_df["residual"])
    thrown_values = merged_df[np.abs(merged_df["residual"]) > cutoff]
    bad_df = pd.concat([bad_df, thrown_values])
    merged_df = merged_df[np.abs(merged_df["residual"]) <= cutoff].copy()

    while not thrown_values.empty:  # runs as long as there are thrown_values

        p0 = tuple(cfw_coefs)  # initialize coefficients with previous results
        weights = calculate_weights(merged_df["CTDPRS"])
        fit_data = tuple(merged_df[v] for v in fit_vars)  # merged_df changes each loop
        res = scipy.optimize.minimize(
            PMEL_oxy_weighted_residual,
            x0=p0,
            args=(weights, fit_data, merged_df["REFOXY"]),
            bounds=[(None, None), (None, None), (0, None), (None, None), (None, None)],
        )

        cfw_coefs = res.x
        merged_df["CTDOXY"] = _PMEL_oxy_eq(cfw_coefs, fit_data)
        merged_df["residual"] = merged_df["REFOXY"] - merged_df["CTDOXY"]
        cutoff = 2.8 * np.std(merged_df["residual"])
        thrown_values = merged_df[np.abs(merged_df["residual"]) > cutoff]
        # TODO: get some kind of logging in here in case things go awry
        # e.g. count of thrown values, start/final stdev, failing to converge, etc.
        bad_df = pd.concat([bad_df, thrown_values])
        merged_df = merged_df[np.abs(merged_df["residual"]) <= cutoff].copy()

    # intermediate plots to diagnose data chunks goodness
    # TODO: implement into bokeh/flask dashboard
    if f_suffix is not None:
        f_out = f"{cfg.fig_dirs['ox']}sbe43_residual{f_suffix}.pdf"
        ctd_plots._intermediate_residual_plot(
            merged_df["residual"],
            merged_df["CTDPRS"],
            merged_df["SSSCC"],
            xlabel="CTDOXY Residual (umol/kg)",
            f_out=f_out,
            xlim=(-10, 10),
        )

    merged_df["CTDOXY_FLAG_W"] = 2
    bad_df["CTDOXY_FLAG_W"] = 3
    df = pd.concat([merged_df, bad_df])

    return cfw_coefs, df


def prepare_oxy(btl_df, time_df, ssscc_list):
    """
    Calculate oxygen-related variables needed for calibration:
    sigma, oxygen solubility (OS), and bottle oxygen

    Parameters
    ----------
    btl_df : DataFrame
        CTD data at bottle stops
    time_df : DataFrame
        Continuous CTD data
    ssscc_list : list of str
        List of stations to process

    Returns
    -------

    """
    # Calculate SA and CT
    btl_df["SA"] = gsw.SA_from_SP(
        btl_df[cfg.column["sal"]],
        btl_df[cfg.column["p"]],
        btl_df[cfg.column["lon"]],
        btl_df[cfg.column["lat"]],
    )
    btl_df["CT"] = gsw.CT_from_t(
        btl_df["SA"],
        btl_df[cfg.column["t1"]],  # oxygen sensor is on primary line (ie t1)
        btl_df[cfg.column["p"]],
    )
    time_df["SA"] = gsw.SA_from_SP(
        time_df[cfg.column["sal"]],
        time_df[cfg.column["p"]],
        time_df[cfg.column["lon"]],
        time_df[cfg.column["lat"]],
    )
    time_df["CT"] = gsw.CT_from_t(
        time_df["SA"],
        time_df[cfg.column["t1"]],  # oxygen sensor is on primary line (ie t1)
        time_df[cfg.column["p"]],
    )

    # calculate sigma
    btl_df["sigma_btl"] = gsw.sigma0(btl_df["SA"], btl_df["CT"])
    time_df["sigma_btl"] = gsw.sigma0(time_df["SA"], time_df["CT"])

    # Calculate oxygen solubility in µmol/kg
    btl_df["OS"] = gsw.O2sol(
        btl_df["SA"],
        btl_df["CT"],
        btl_df[cfg.column["p"]],
        btl_df[cfg.column["lon"]],
        btl_df[cfg.column["lat"]],
    )
    time_df["OS"] = gsw.O2sol(
        time_df["SA"],
        time_df["CT"],
        time_df[cfg.column["p"]],
        time_df[cfg.column["lon"]],
        time_df[cfg.column["lat"]],
    )
    # Convert CTDOXY units
    btl_df["CTDOXY"] = oxy_ml_to_umolkg(btl_df["CTDOXY1"], btl_df["sigma_btl"])
    # Calculate bottle oxygen
    btl_df[cfg.column["refO"]] = calculate_bottle_oxygen(
        ssscc_list,
        btl_df["SSSCC"],
        btl_df["TITR_VOL"],
        btl_df["TITR_TEMP"],
        btl_df["FLASKNO"],
    )
    btl_df[cfg.column["refO"]] = oxy_ml_to_umolkg(
        btl_df[cfg.column["refO"]], btl_df["sigma_btl"]
    )
    btl_df["OXYGEN_FLAG_W"] = flagging.nan_values(btl_df[cfg.column["refO"]])
    # Load manual OXYGEN flags
    if Path("data/oxygen/manual_oxy_flags.csv").exists():
        manual_flags = pd.read_csv(
            "data/oxygen/manual_oxy_flags.csv", dtype={"SSSCC": str}
        )
        for _, flags in manual_flags.iterrows():
            df_row = (btl_df["SSSCC"] == flags["SSSCC"]) & (
                btl_df["btl_fire_num"] == flags["SAMPNO"]
            )
            btl_df.loc[df_row, "OXYGEN_FLAG_W"] = flags["Flag"]

    return True


def calibrate_oxy(btl_df, time_df, ssscc_list):
    """
    Non-linear least squares fit chemical sensor oxygen against bottle oxygen.

    Parameters
    ----------
    btl_df : DataFrame
        CTD data at bottle stops
    time_df : DataFrame
        Continuous CTD data
    ssscc_list : list of str
        List of stations to process

    Returns
    -------

    """
    log.info("Calibrating oxygen (SBE43)")
    # Plot all pre fit data
    f_out = f"{cfg.fig_dirs['ox']}sbe43_residual_all_prefit.pdf"
    ctd_plots._intermediate_residual_plot(
        btl_df["OXYGEN"] - btl_df["CTDOXY"],
        btl_df["CTDPRS"],
        btl_df["SSSCC"],
        xlabel="CTDOXY Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )
    # Prep vars, dfs, etc.
    all_sbe43_merged = pd.DataFrame()
    sbe43_dict = {}
    all_sbe43_fit = pd.DataFrame()

    btl_df["dv_dt"] = np.nan  # initialize column
    # Density match time/btl oxy dataframes
    for ssscc in ssscc_list:
        time_data = time_df[time_df["SSSCC"] == ssscc].copy()
        btl_data = btl_df[btl_df["SSSCC"] == ssscc].copy()
        # can't calibrate without bottle oxygen ("OXYGEN")
        if (btl_data["OXYGEN_FLAG_W"] == 9).all():
            sbe43_dict[ssscc] = np.full(5, np.nan)
            log.warning(ssscc + " skipped, all oxy data is NaN")
            continue
        sbe43_merged = match_sigmas(
            btl_data[cfg.column["p"]],
            btl_data[cfg.column["refO"]],
            btl_data["CTDTMP1"],
            btl_data["SA"],
            time_data["OS"],
            time_data[cfg.column["p"]],
            time_data[cfg.column["t1"]],
            time_data["SA"],
            time_data[cfg.column["oxyvolts"]],
            time_data["scan_datetime"],
        )
        sbe43_merged = sbe43_merged.reindex(btl_data.index)  # add nan rows back in
        btl_df.loc[
            btl_df["SSSCC"] == ssscc, ["CTDOXYVOLTS", "dv_dt", "OS"]
        ] = sbe43_merged[["CTDOXYVOLTS", "dv_dt", "OS"]]
        sbe43_merged["SSSCC"] = ssscc
        all_sbe43_merged = pd.concat([all_sbe43_merged, sbe43_merged])
        log.info(ssscc + " density matching done")

    # Only fit using OXYGEN flagged good (2)
    all_sbe43_merged = all_sbe43_merged[btl_df["OXYGEN_FLAG_W"] == 2].copy()

    # Fit ALL oxygen stations together to get initial coefficient guess
    (sbe_coef0, _) = sbe43_oxy_fit(all_sbe43_merged, f_suffix="_ox0")
    sbe43_dict["ox0"] = sbe_coef0

    # Fit each cast individually
    for ssscc in ssscc_list:
        sbe_coef, sbe_df = sbe43_oxy_fit(
            all_sbe43_merged.loc[all_sbe43_merged["SSSCC"] == ssscc].copy(),
            sbe_coef0=sbe_coef0,
            f_suffix=f"_{ssscc}",
        )
        # build coef dictionary
        if ssscc not in sbe43_dict.keys():  # don't overwrite NaN'd stations
            sbe43_dict[ssscc] = sbe_coef
        # all non-NaN oxygen data with flags
        all_sbe43_fit = pd.concat([all_sbe43_fit, sbe_df])

    # TODO: save outlier data from fits?
    # TODO: secondary oxygen flagging step (instead of just taking outliers from fit routine)

    # apply coefs
    time_df["CTDOXY"] = np.nan
    for ssscc in ssscc_list:
        if np.isnan(sbe43_dict[ssscc]).all():
            log.warning(
                f"{ssscc} missing oxy data, leaving nan values and flagging as 9"
            )
            time_df.loc[time_df["SSSCC"] == ssscc, "CTDOXY_FLAG_W"] = 9
            continue
        btl_rows = (btl_df["SSSCC"] == ssscc).values
        time_rows = (time_df["SSSCC"] == ssscc).values
        btl_df.loc[btl_rows, "CTDOXY"] = _PMEL_oxy_eq(
            sbe43_dict[ssscc],
            (
                btl_df.loc[btl_rows, cfg.column["oxyvolts"]],
                btl_df.loc[btl_rows, cfg.column["p"]],
                btl_df.loc[btl_rows, cfg.column["t1"]],
                btl_df.loc[btl_rows, "dv_dt"],
                btl_df.loc[btl_rows, "OS"],
            ),
        )
        log.info(ssscc + " btl data fitting done")
        time_df.loc[time_rows, "CTDOXY"] = _PMEL_oxy_eq(
            sbe43_dict[ssscc],
            (
                time_df.loc[time_rows, cfg.column["oxyvolts"]],
                time_df.loc[time_rows, cfg.column["p"]],
                time_df.loc[time_rows, cfg.column["t1"]],
                time_df.loc[time_rows, "dv_dt"],
                time_df.loc[time_rows, "OS"],
            ),
        )
        log.info(ssscc + " time data fitting done")

    # flag CTDOXY with more than 1% difference
    time_df["CTDOXY_FLAG_W"] = 2  # TODO: actual flagging of some kind?
    btl_df["CTDOXY_FLAG_W"] = flagging.by_percent_diff(
        btl_df["CTDOXY"], btl_df["OXYGEN"], percent_thresh=1
    )

    # Plot all post fit data
    f_out = f"{cfg.fig_dirs['ox']}sbe43_residual_all_postfit.pdf"
    ctd_plots._intermediate_residual_plot(
        btl_df["OXYGEN"] - btl_df["CTDOXY"],
        btl_df["CTDPRS"],
        btl_df["SSSCC"],
        xlabel="CTDOXY Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )
    f_out = f"{cfg.fig_dirs['ox']}sbe43_residual_all_postfit_flag2.pdf"
    flag2 = btl_df["CTDOXY_FLAG_W"] == 2
    ctd_plots._intermediate_residual_plot(
        btl_df.loc[flag2, "OXYGEN"] - btl_df.loc[flag2, "CTDOXY"],
        btl_df.loc[flag2, "CTDPRS"],
        btl_df.loc[flag2, "SSSCC"],
        xlabel="CTDOXY Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )

    # export fitting coefs
    sbe43_coefs = pd.DataFrame.from_dict(
        sbe43_dict, orient="index", columns=["Soc", "Voffset", "Tau20", "Tcorr", "E"]
    ).applymap(lambda x: np.format_float_scientific(x, precision=4, exp_digits=1))
    sbe43_coefs.to_csv(cfg.dirs["logs"] + "sbe43_coefs.csv")

    return True
